# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for models constructed via global_circulation.ml.model_builder.py."""

import functools
from absl.testing import absltest
from absl.testing import parameterized
from dinosaur import coordinate_systems
from dinosaur import layer_coordinates
from dinosaur import primitive_equations
from dinosaur import primitive_equations_states
from dinosaur import scales
from dinosaur import shallow_water
from dinosaur import shallow_water_states
from dinosaur import sigma_coordinates
from dinosaur import spherical_harmonic
from dinosaur import typing
from dinosaur import xarray_utils
import gin
import haiku as hk
import jax
import jax.numpy as jnp
from neuralgcm import model_builder
from neuralgcm import physics_specifications
import numpy as np

units = scales.units


class ShallowWaterTest(parameterized.TestCase):
  """Tests that ML-configured ShallowWater solver produces expected shapes."""

  def setUp(self):
    """Initializes gin-configs used in tests."""
    super().setUp()
    gin.enter_interactive_mode()
    self.physics_gin_config = '\n'.join([
        'get_physics_specs.construct_fn = @shallow_water_specs_constructor',
        'shallow_water_specs_constructor.density_vals = [1.]',
    ])
    self.model_gin_config = '\n'.join([
        'WhirlModel.model_cls = @ModularStepModel',
        'ModularStepModel.advance_module = @EquationStep',
        'ModularStepModel.encoder_module = %ENCODER_MODULE',
        'ModularStepModel.decoder_module = %DECODER_MODULE',
        'EquationStep.time_integrator = %INTEGRATOR_MODULE',
        'EquationStep.filter_modules = %FILTERS',
        'EquationStep.equation_module = @ShallowWaterEquations',
    ])

  @parameterized.parameters(
      # TODO(pnorgaard) Reenable after making forcing work with leapfrog
      # dict(encoder_module='@ShallowWaterLeapfrogEncoder',
      #      integrator_module='@semi_implicit_leapfrog',
      #      decoder_module='@LeapfrogSliceDictDecoder',
      #      filter_modules='(@ExponentialLeapfrogFilter,)',
      #      model_state_fn=lambda x: (x, x)),
      dict(
          encoder_module='@ShallowWaterStateEncoder',
          integrator_module='@imex_rk_sil3',
          decoder_module='@StateToDictDecoder',
          filter_modules='(@ExponentialFilter,)',
          model_state_fn=lambda x: x,
      ),
  )
  def test_shapes_of_model_function(
      self,
      encoder_module,
      integrator_module,
      decoder_module,
      filter_modules,
      model_state_fn,  # constructs expected model state from `State` snapshot.
  ):
    """Tests output shapes of encode, advance, decode, trajectory functions."""
    gin_bindings = '\n'.join([
        f'ENCODER_MODULE = {encoder_module}',
        f'INTEGRATOR_MODULE = {integrator_module}',
        f'DECODER_MODULE = {decoder_module}',
        f'FILTERS = {filter_modules}',
    ])
    gin.clear_config()
    gin.parse_config(self.physics_gin_config)
    gin.parse_config(self.model_gin_config)
    gin.parse_config(gin_bindings)

    grid = spherical_harmonic.Grid.with_wavenumbers(32)
    vertical_grid = layer_coordinates.LayerCoordinates(1)
    coords = coordinate_systems.CoordinateSystem(grid, vertical_grid)
    dt = 1e-3
    physics_specs = physics_specifications.get_physics_specs()

    initial_state_fn, aux_features = (
        shallow_water_states.barotropic_instability_tc(coords, physics_specs)
    )
    instant_state = initial_state_fn(jax.random.PRNGKey(42))
    model = model_builder.WhirlModel(coords, dt, physics_specs, aux_features)

    # model state is a pair of instant states for leapfrog integration.
    state = model_state_fn(instant_state)
    model_state = typing.ModelState(state)
    # data state is an instantenious configuration;
    # decoder should map model_state to data_state;
    data_state = instant_state.asdict()
    # state and data trajectories have extra time dimensions;
    stack_fn = lambda *args: jnp.stack(args)
    traj_length = 3
    state_trajectory = jax.tree_util.tree_map(
        stack_fn, *([model_state] * traj_length)
    )
    input_length = 2  # encoder takes trajectory as input for assimilation.
    data_trajectory = jax.tree_util.tree_map(
        stack_fn, *([data_state] * input_length)
    )
    forcing_data = None
    forcing = None

    rng_key = jax.random.PRNGKey(42)
    params = model.init_params(rng_key, data_trajectory, forcing_data)

    get_shape_fn = lambda tree: jax.tree_util.tree_map(np.shape, tree)

    with self.subTest('encoder'):
      actual = jax.tree_util.tree_leaves(
          model.encode_fn(params, rng_key, data_trajectory, forcing).state
      )
      expected = jax.tree_util.tree_leaves(model_state.state)
      self.assertSameStructure(get_shape_fn(actual), get_shape_fn(expected))

    with self.subTest('decoder'):
      actual = jax.tree_util.tree_leaves(
          model.decode_fn(params, rng_key, model_state, forcing)
      )
      expected = jax.tree_util.tree_leaves(data_state)
      self.assertSameStructure(get_shape_fn(actual), get_shape_fn(expected))

    with self.subTest('advance'):
      actual = jax.tree_util.tree_leaves(
          model.advance_fn(params, rng_key, model_state, forcing).state
      )
      expected = jax.tree_util.tree_leaves(model_state.state)
      self.assertSameStructure(get_shape_fn(actual), get_shape_fn(expected))

    with self.subTest('trajectory'):

      def trajectory_fwd(x, length, forcing_data):
        return model.model_cls().trajectory(
            x, length, forcing_data=forcing_data
        )

      trajectory_model = hk.transform(trajectory_fwd)
      actual = jax.tree_util.tree_leaves(
          trajectory_model.apply(
              params, rng_key, model_state, traj_length, forcing_data
          )
      )
      expected = jax.tree_util.tree_leaves((model_state, state_trajectory))
      self.assertSameStructure(get_shape_fn(actual), get_shape_fn(expected))

  def test_equivalence_with_explicit_integration(self):
    """Tests that configured model produces identical results to `integrate`."""
    leapfrog_gin_bindings = '\n'.join([
        'ENCODER_MODULE = @ShallowWaterLeapfrogEncoder',
        'INTEGRATOR_MODULE = @semi_implicit_leapfrog',
        'DECODER_MODULE = @LeapfrogSliceDecoder',
        'FILTERS = (@ExponentialLeapfrogFilter, @RobertAsselinLeapfrogFilter)',
    ])
    gin.clear_config()
    gin.parse_config(self.physics_gin_config)
    gin.parse_config(self.model_gin_config)
    gin.parse_config(leapfrog_gin_bindings)

    grid = spherical_harmonic.Grid.with_wavenumbers(32)
    vertical_grid = layer_coordinates.LayerCoordinates(1)
    coords = coordinate_systems.CoordinateSystem(grid, vertical_grid)
    physics_specs = physics_specifications.get_physics_specs()
    dt = physics_specs.nondimensionalize(30.0 * units.s)

    initial_state_fn, aux_features = (
        shallow_water_states.barotropic_instability_tc(coords, physics_specs)
    )
    instant_state = initial_state_fn(jax.random.PRNGKey(42))

    initial_state = (instant_state, instant_state)
    initial_model_state = typing.ModelState(initial_state)
    trajectory_length = 30
    # configured model prediction.
    model = model_builder.WhirlModel(coords, dt, physics_specs, aux_features)
    trajectory_fwd = lambda x: model.model_cls().trajectory(  # pylint: disable=g-long-lambda
        x, trajectory_length, forcing_data={}
    )
    trajectory_model = hk.without_apply_rng(hk.transform(trajectory_fwd))
    key = jax.random.PRNGKey(42)
    params = trajectory_model.init(key, initial_model_state)
    trajectory_fn = jax.jit(trajectory_model.apply)
    actual_traj = trajectory_fn(params, initial_model_state)[1]
    # prediction using `shallow_water_leapfrog_trajectory` function.
    explicit_trajectory_fn = shallow_water.shallow_water_leapfrog_trajectory(
        coords=coords,
        dt=dt,
        physics_specs=physics_specs,
        mean_potential=aux_features[xarray_utils.REF_POTENTIAL_KEY],
        inner_steps=1,
        outer_steps=trajectory_length,
        filters=shallow_water.default_filters(grid, dt),
    )
    explicit_trajectory_fn = jax.jit(explicit_trajectory_fn)

    _, expected_traj = explicit_trajectory_fn(initial_state)
    for actual, expected in zip(
        jax.tree_util.tree_leaves(actual_traj),
        jax.tree_util.tree_leaves(expected_traj),
    ):
      np.testing.assert_allclose(actual, expected, atol=5e-6)


class PrimitiveEquationsTest(parameterized.TestCase):
  """Tests ML-configured primitive equation solver."""

  def setUp(self):
    """Initializes gin-configs and initial conditions used in tests."""
    super().setUp()
    gin.enter_interactive_mode()
    self.physics_gin_config = '\n'.join([
        'get_physics_specs.construct_fn = @primitive_eq_specs_constructor',
    ])
    gin.clear_config()
    gin.parse_config(self.physics_gin_config)
    self.physics_specs = physics_specifications.get_physics_specs()
    # setting up coordinates and initial state.
    grid = spherical_harmonic.Grid.with_wavenumbers(32)
    vertical_grid = sigma_coordinates.SigmaCoordinates.equidistant(4)
    self.coords = coordinate_systems.CoordinateSystem(grid, vertical_grid)
    initial_state_fn, aux_features = (
        primitive_equations_states.isothermal_rest_atmosphere(
            self.coords, self.physics_specs
        )
    )
    self.initial_instant_state = initial_state_fn(jax.random.PRNGKey(42))
    self.aux_features = aux_features
    # setting up base model configuration.
    diagnostic_state = primitive_equations.compute_diagnostic_state(
        self.initial_instant_state, self.coords
    )
    nodal_state = jax.tree_util.tree_map(
        grid.to_nodal, self.initial_instant_state
    )
    inputs_mean = jax.tree_util.tree_map(
        lambda x: float(np.mean(x)), diagnostic_state.asdict()
    )
    inputs_std = jax.tree_util.tree_map(
        lambda x: 0.1 + float(np.std(x)), diagnostic_state.asdict()
    )
    outputs_mean = jax.tree_util.tree_map(
        lambda x: float(np.mean(x)), nodal_state.asdict()
    )
    outputs_std = jax.tree_util.tree_map(
        lambda x: 0.1 + float(np.std(x)), nodal_state.asdict()
    )
    self.model_gin_config = '\n'.join([
        'WhirlModel.model_cls = @ModularStepModel',
        'ModularStepModel.advance_module = @EquationStep',
        'ModularStepModel.encoder_module = %ENCODER_MODULE',
        'ModularStepModel.decoder_module = %DECODER_MODULE',
        'EquationStep.time_integrator = %INTEGRATOR_MODULE',
        'EquationStep.filter_modules = %FILTERS',
        'EquationStep.equation_module = %EQUATION_MODULE',
        # configuring a basic values for ML modules
        (
            'composed_equations_module.equation_modules = '
            '(@PrimitiveEquations, @DirectNeuralEquations)'
        ),
        (
            'with_vertical_diffusion/composed_equations_module.equation_modules'
            ' = (@PrimitiveEquations, @VerticalDiffusion)'
        ),
        'VerticalDiffusion.timescale = "3 hours"',
        (
            'DirectNeuralEquations.modal_to_nodal_features_module = '
            '@PrimitiveEquationsDiagnosticState'
        ),
        (
            'PrimitiveEquationsDiagnosticState.features_transform_module = '
            '@inputs/ShiftAndNormalize'
        ),
        (
            'DirectNeuralEquations.tendency_transform_module = '
            '@outputs/ShiftAndNormalize'
        ),
        'DirectNeuralEquations.nodal_mapping_module = @NodalMapping',
        'NodalMapping.tower_factory = @ColumnTower',
        'ColumnTower.column_net_factory = @MlpUniform',
        'MlpUniform.num_hidden_units = 32',
        'MlpUniform.num_hidden_layers = 2',
        f'inputs/ShiftAndNormalize.shifts = {inputs_mean}',
        f'inputs/ShiftAndNormalize.scales = {inputs_std}',
        f'outputs/ShiftAndNormalize.shifts = {outputs_mean}',
        f'outputs/ShiftAndNormalize.scales = {outputs_std}',
    ])

  @parameterized.parameters(
      # TODO(pnorgaard) Reenable after making forcing work with leapfrog
      # dict(encoder_module='@PrimitiveEquationLeapfrogEncoder',
      #      equation_module='@PrimitiveEquations',
      #      integrator_module='@semi_implicit_leapfrog',
      #      decoder_module='@LeapfrogSliceDictDecoder',
      #      filter_modules='(@ExponentialLeapfrogFilter,)',
      #      model_state_fn=lambda x: (x, x)),
      dict(
          encoder_module='@PrimitiveEquationStateEncoder',
          equation_module='@PrimitiveEquations',
          integrator_module='@imex_rk_sil3',
          decoder_module='@StateToDictDecoder',
          filter_modules='(@ExponentialFilter,)',
          model_state_fn=lambda x: x,
      ),
      dict(
          encoder_module='@PrimitiveEquationStateEncoder',
          equation_module='@composed_equations_module',
          integrator_module='@imex_rk_sil3',
          decoder_module='@StateToDictDecoder',
          filter_modules='(@ExponentialFilter,)',
          model_state_fn=lambda x: x,
      ),
      dict(
          encoder_module='@PrimitiveEquationStateEncoder',
          equation_module='@with_vertical_diffusion/composed_equations_module',
          integrator_module='@imex_rk_sil3',
          decoder_module='@StateToDictDecoder',
          filter_modules='(@ExponentialFilter,)',
          model_state_fn=lambda x: x,
      ),
  )
  def test_shapes_of_model_function(
      self,
      encoder_module,
      equation_module,
      integrator_module,
      decoder_module,
      filter_modules,
      model_state_fn,  # constructs expected model state from `State` snapshot.
  ):
    """Tests output shapes of encode, advance, decode, trajectory functions."""
    gin_bindings = '\n'.join([
        f'ENCODER_MODULE = {encoder_module}',
        f'EQUATION_MODULE = {equation_module}',
        f'INTEGRATOR_MODULE = {integrator_module}',
        f'DECODER_MODULE = {decoder_module}',
        f'FILTERS = {filter_modules}',
    ])
    gin.clear_config()
    gin.parse_config(self.model_gin_config)
    gin.parse_config(gin_bindings)
    dt = self.physics_specs.nondimensionalize(30 * units.s)
    model = model_builder.WhirlModel(
        self.coords, dt, self.physics_specs, self.aux_features
    )

    # model state is a pair of instant states for leapfrog integration.
    state = model_state_fn(self.initial_instant_state)
    model_state = typing.ModelState(state)
    # data state is an instantenious configuration;
    # decoder should map model_state to data_state;
    data_state = self.initial_instant_state.asdict()
    # state and data trajectories have extra time dimensions;
    stack_fn = lambda *args: jnp.stack(args)
    traj_length = 3
    state_trajectory = jax.tree_util.tree_map(
        stack_fn, *([model_state] * traj_length)
    )
    input_length = 2  # encoder takes trajectory as input for assimilation.
    data_trajectory = jax.tree_util.tree_map(
        stack_fn, *([data_state] * input_length)
    )
    forcing_data = None
    forcing = None

    rng_key = jax.random.PRNGKey(42)
    params = model.init_params(rng_key, data_trajectory, forcing_data)
    get_shape_fn = lambda tree: jax.tree_util.tree_map(np.shape, tree)

    with self.subTest('encoder'):
      actual = jax.tree_util.tree_leaves(
          model.encode_fn(params, rng_key, data_trajectory, forcing).state
      )
      expected = jax.tree_util.tree_leaves(model_state.state)
      self.assertSameStructure(get_shape_fn(actual), get_shape_fn(expected))

    with self.subTest('decoder'):
      actual = jax.tree_util.tree_leaves(
          model.decode_fn(params, rng_key, model_state, forcing)
      )
      expected = jax.tree_util.tree_leaves(data_state)
      self.assertSameStructure(get_shape_fn(actual), get_shape_fn(expected))

    with self.subTest('advance'):
      actual = jax.tree_util.tree_leaves(
          model.advance_fn(params, rng_key, model_state, forcing).state
      )
      expected = jax.tree_util.tree_leaves(model_state.state)
      self.assertSameStructure(get_shape_fn(actual), get_shape_fn(expected))

    with self.subTest('trajectory'):

      def trajectory_fwd(x, length):
        return model.model_cls().trajectory(x, length, forcing_data={})

      trajectory_model = hk.transform(trajectory_fwd)
      params = trajectory_model.init(rng_key, model_state, length=traj_length)
      actual = jax.tree_util.tree_leaves(
          trajectory_model.apply(
              params, rng_key, model_state, length=traj_length
          )
      )
      expected = jax.tree_util.tree_leaves((model_state, state_trajectory))
      self.assertSameStructure(get_shape_fn(actual), get_shape_fn(expected))


class NodalModalEncoderDecoderTest(parameterized.TestCase):
  """Tests nodal to nodal encode->decode roundtrip using configured modules."""

  def setUp(self):
    """Defines gin-configs used in tests."""
    super().setUp()
    gin.enter_interactive_mode()
    self.physics_gin_config = '\n'.join([
        'get_physics_specs.construct_fn = @primitive_eq_specs_constructor',
    ])
    encoder_module_name = 'PrimitiveEquationStateEncoder'
    self.model_gin_config = '\n'.join([
        'WhirlModel.model_cls = @ModularStepModel',
        'ModularStepModel.advance_module = @EquationStep',
        f'ModularStepModel.encoder_module = @{encoder_module_name}',
        'ModularStepModel.decoder_module = @StateToDictDecoder',
        f'{encoder_module_name}.transform_module = @InputNodalToModalTransform',
        'StateToDictDecoder.transform_module = @OutputModalToNodalTransform',
        'EquationStep.equation_module = @PrimitiveEquations',
    ])

  @parameterized.parameters(
      dict(
          input_grid=spherical_harmonic.Grid.with_wavenumbers(64),
          model_grid=spherical_harmonic.Grid.with_wavenumbers(64),
          output_grid=spherical_harmonic.Grid.with_wavenumbers(
              32, latitude_spacing='equiangular'
          ),
      ),
      dict(
          input_grid=spherical_harmonic.Grid.with_wavenumbers(
              64, latitude_spacing='equiangular'
          ),
          model_grid=spherical_harmonic.Grid.with_wavenumbers(64),
          output_grid=spherical_harmonic.Grid.with_wavenumbers(
              32, latitude_spacing='equiangular'
          ),
      ),
  )
  def test_nodal_to_nodal_trip(
      self,
      input_grid: spherical_harmonic.Grid,
      model_grid: spherical_harmonic.Grid,
      output_grid: spherical_harmonic.Grid,
  ):
    """Tests that models data->state->data trip produces expected results."""
    gin.clear_config()
    gin.parse_config(self.physics_gin_config)
    gin.parse_config(self.model_gin_config)
    physics_specs = physics_specifications.get_physics_specs()
    # Define coordinates for the model, and for the input and output data.
    vertical_coordinates = sigma_coordinates.SigmaCoordinates.equidistant(4)
    input_coords = coordinate_systems.CoordinateSystem(
        input_grid, vertical_coordinates
    )
    model_coords = coordinate_systems.CoordinateSystem(
        model_grid, vertical_coordinates
    )
    output_coords = coordinate_systems.CoordinateSystem(
        output_grid, vertical_coordinates
    )
    # Generate an initial state to act as the input data.
    initial_state_fn, aux_features = primitive_equations_states.steady_state_jw(
        input_coords, physics_specs
    )
    initial_modal_state = initial_state_fn()
    forcing_data = None
    forcing = None
    data_inputs = jax.tree_util.tree_map(
        lambda x: np.expand_dims(input_grid.to_nodal(x), 0),
        initial_modal_state.asdict(),
    )
    model = model_builder.WhirlModel(
        model_coords,
        1.0,
        physics_specs,
        aux_features,
        input_coords=input_coords,
        output_coords=output_coords,
    )
    # to initialize we pass a data trajectory.
    rng = jax.random.PRNGKey(42)
    params = model.init_params(rng, data_inputs, forcing_data)

    with self.subTest('model_state'):
      # check that `encode` produces an expected state.
      model_state = jax.jit(model.encode_fn)(params, rng, data_inputs, forcing)
      actual_prognostic_vars = model_state.state
      jax.tree.map(
          functools.partial(np.testing.assert_allclose, atol=1e-4),
          actual_prognostic_vars,
          initial_modal_state,
      )

    with self.subTest('decoded_state'):
      # check that `decode(encode(x))` is close to identity.
      target_lon_k, target_lat_k = output_grid.modal_shape
      target_modal_state = jax.tree_util.tree_map(
          lambda x: x[:, :target_lon_k, :target_lat_k],
          initial_modal_state.asdict(),
      )
      target_nodal = jax.tree_util.tree_map(
          output_grid.to_nodal, target_modal_state
      )

      def encode_decode_fn(p, x, forcing):
        x = model.encode_fn(p, rng, x, forcing)
        return model.decode_fn(p, rng, x, forcing)

      encode_decode_fn = jax.jit(encode_decode_fn)
      actual_nodal = encode_decode_fn(params, data_inputs, forcing)
      jax.tree.map(
          functools.partial(np.testing.assert_allclose, atol=2e-4),
          actual_nodal,
          target_nodal,
      )


if __name__ == '__main__':
  jax.config.parse_flags_with_absl()
  absltest.main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines configurable time integrators to be used by models."""

from typing import Callable
from dinosaur import time_integration
from dinosaur import typing
import gin


TimeIntegrator = Callable[
    [time_integration.ImplicitExplicitODE, typing.Numeric], typing.TimeStepFn]


backward_forward_euler = gin.external_configurable(
    time_integration.backward_forward_euler)
crank_nicolson_rk2 = gin.external_configurable(
    time_integration.crank_nicolson_rk2)
crank_nicolson_rk3 = gin.external_configurable(
    time_integration.crank_nicolson_rk3)
crank_nicolson_rk4 = gin.external_configurable(
    time_integration.crank_nicolson_rk4)
imex_rk_sil3 = gin.external_configurable(time_integration.imex_rk_sil3)
semi_implicit_leapfrog = gin.external_configurable(
    time_integration.semi_implicit_leapfrog)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Loads datasets."""

import functools
import itertools
import json
import logging
import math
import multiprocessing
import random
from typing import Any, Callable, Iterator, Mapping, Optional, Tuple

import jax
import numpy as np
import pandas as pd
import tensorflow.compat.v2 as tf
import xarray


Pytree = Any
# pylint: disable=g-bare-generic
# pylint: disable=logging-fstring-interpolation


def drop_static_vars(dataset: xarray.Dataset) -> xarray.Dataset:
  """Drop fields that are static and do not vary with time."""
  has_sample_dim = 'sample' in dataset.coords
  vars_to_drop = []
  for name, var in dataset.items():
    if 'time' not in var.dims:
      vars_to_drop.append(name)
    elif has_sample_dim and var.dims[:2] != ('sample', 'time'):
      raise ValueError(f'dimensions for variable {name} do not start with '
                       f"'sample' and 'time': {var.dims}")
    elif not has_sample_dim and var.dims[0] != 'time':
      raise ValueError(f'dimensions for variable {name} do not start with '
                       f"'time': {var.dims}")
  return dataset.drop_vars(vars_to_drop)


def attrs_from_dataset(
    dataset: xarray.Dataset,
    time_series_length: int,
    subsample_rate: int = 1,
) -> dict:
  """Extracts attributes from `dataset`."""
  attrs = dict(dataset.attrs)
  attrs['trajectory_length'] = time_series_length
  attrs['time_subsample_rate'] = subsample_rate
  delta_t = (dataset.time[1] - dataset.time[0]).data
  if not np.issubdtype(dataset.time.dtype, np.floating):
    logging.info(f'converting non-float {delta_t=} to seconds')
    delta_t = np.timedelta64(delta_t, 's') / np.timedelta64(1, 's')
    attrs['save_dt_units'] = 's'
  else:
    attrs['save_dt_units'] = 'dimensionless'
  attrs['save_dt'] = float(delta_t) * subsample_rate
  return attrs



# Copyright 2024 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Handling conversion dimensional <-> nondimensional quantities."""

from __future__ import annotations

import dataclasses

from neuralgcm.experimental import scales
from neuralgcm.experimental import typing
import numpy as np


# pylint: disable=invalid-name  # for standard physical constant names.

Numeric = typing.Numeric
Quantity = typing.Quantity


parse_units = scales.parse_units


@dataclasses.dataclass(frozen=True)
class SimUnits:
  """Provides units & time conversion methods and common constants."""

  radius: float
  angular_velocity: float
  gravity_acceleration: float
  ideal_gas_constant: float
  water_vapor_gas_constant: float
  water_vapor_isobaric_heat_capacity: float
  kappa: float
  reference_datetime: np.datetime64
  scale: scales.Scale

  @property
  def R(self) -> float:
    """Alias for `ideal_gas_constant`."""
    return self.ideal_gas_constant

  @property
  def R_vapor(self) -> float:
    """Alias for `ideal_gas_constant`."""
    return self.water_vapor_gas_constant

  @property
  def g(self) -> float:
    """Alias for `gravity_acceleration`."""
    return self.gravity_acceleration

  @property
  def Cp(self) -> float:
    """Isobaric heat capacity."""
    return self.ideal_gas_constant / self.kappa

  @property
  def Cp_vapor(self) -> float:
    """Alias for `water_vapor_isobaric_heat_capacity`."""
    return self.water_vapor_isobaric_heat_capacity

  def nondimensionalize_timedelta64(self, timedelta: np.timedelta64) -> Numeric:
    """Non-dimensionalizes and rescales a numpy timedelta."""
    base_unit = 's'
    return self.scale.nondimensionalize(
        timedelta / np.timedelta64(1, base_unit) * typing.units(base_unit)
    )

  def dimensionalize_timedelta64(self, value: Numeric) -> np.timedelta64:
    """Rescales and casts the given non-dimensional value to timedelta64."""
    base_unit = 's'  # return value is rounded down to nearest base_unit
    dt = self.scale.dimensionalize(value, typing.units(base_unit)).m
    if isinstance(dt, np.ndarray):
      return dt.astype(f'timedelta64[{base_unit}]')
    else:
      return np.timedelta64(int(dt), base_unit)

  def datetime64_to_sim_time(self, datetime64: np.ndarray) -> typing.Numeric:
    """Converts a datetime64 array to sim_time."""
    return self.scale.nondimensionalize(
        ((datetime64 - self.reference_datetime) / np.timedelta64(1, 'h'))
        * typing.units.hour
    )

  def sim_time_to_datetime64(self, sim_time: np.ndarray) -> np.ndarray:
    """Converts a sim_time array to datetime64."""
    minutes = self.scale.dimensionalize(sim_time, typing.units.minute).magnitude
    delta = np.array(np.round(minutes).astype(int), 'timedelta64[m]')
    return self.reference_datetime + delta

  def nondimensionalize(
      self, quantity: typing.Quantity
  ) -> typing.Numeric:
    """Non-dimensionalizes and rescales `quantity`."""
    return self.scale.nondimensionalize(quantity)

  def dimensionalize(
      self,
      value: typing.Numeric,
      unit: typing.units.Unit,
      as_quantity: bool = True,
  ):
    """Rescales and adds units to the given non-dimensional value."""
    dimensionalized = self.scale.dimensionalize(value, unit)
    return dimensionalized if as_quantity else dimensionalized.m

  @classmethod
  def from_si(
      cls,
      radius_si: Quantity = scales.RADIUS,
      angular_velocity_si: Quantity = scales.ANGULAR_VELOCITY,
      gravity_acceleration_si: Quantity = scales.GRAVITY_ACCELERATION,
      ideal_gas_constant_si: Quantity = scales.IDEAL_GAS_CONSTANT,
      water_vapor_gas_constant_si: Quantity = scales.IDEAL_GAS_CONSTANT_H20,
      water_vapor_isobaric_heat_capacity_si: Quantity = scales.WATER_VAPOR_CP,
      kappa_si: Quantity = scales.KAPPA,
      reference_datetime: np.datetime64 = np.datetime64('1979-01-01T00:00:00'),
      scale: scales.Scale = scales.DEFAULT_SCALE,
  ) -> SimUnits:
    """Constructs `SimUnits` from SI constants."""
    return cls(
        scale.nondimensionalize(radius_si),
        scale.nondimensionalize(angular_velocity_si),
        scale.nondimensionalize(gravity_acceleration_si),
        scale.nondimensionalize(ideal_gas_constant_si),
        scale.nondimensionalize(water_vapor_gas_constant_si),
        scale.nondimensionalize(water_vapor_isobaric_heat_capacity_si),
        scale.nondimensionalize(kappa_si),
        reference_datetime=reference_datetime,
        scale=scale,
    )


DEFAULT_UNITS = SimUnits.from_si(
    radius_si=scales.RADIUS,
    angular_velocity_si=scales.ANGULAR_VELOCITY,
    gravity_acceleration_si=scales.GRAVITY_ACCELERATION,
    ideal_gas_constant_si=scales.IDEAL_GAS_CONSTANT,
    water_vapor_gas_constant_si=scales.IDEAL_GAS_CONSTANT_H20,
    water_vapor_isobaric_heat_capacity_si=scales.WATER_VAPOR_CP,
    kappa_si=scales.KAPPA,
    scale=scales.ATMOSPHERIC_SCALE,
    reference_datetime=np.datetime64('1979-01-01T00:00:00'),
)

ONE_MINUTE_DT_SCALE = scales.Scale(
    scales.RADIUS,
    typing.units('1 minute'),
    scales.MASS_OF_DRY_ATMOSPHERE,  # mass
    1 * typing.units.degK,
)

ONE_MINUTE_DT_UNITS = SimUnits.from_si(
    radius_si=scales.RADIUS,
    angular_velocity_si=scales.ANGULAR_VELOCITY,
    gravity_acceleration_si=scales.GRAVITY_ACCELERATION,
    ideal_gas_constant_si=scales.IDEAL_GAS_CONSTANT,
    water_vapor_gas_constant_si=scales.IDEAL_GAS_CONSTANT_H20,
    water_vapor_isobaric_heat_capacity_si=scales.WATER_VAPOR_CP,
    kappa_si=scales.KAPPA,
    scale=ONE_MINUTE_DT_SCALE,
    reference_datetime=np.datetime64('1979-01-01T00:00:00'),
)


def maybe_nondimensionalize(
    x: typing.Numeric | typing.Quantity | str | None,
    sim_units: SimUnits,
) -> None | typing.Numeric:
  """Calls nondimensionalize on Quantity or str, otherwise passthrough."""
  if x is None:
    return None
  elif isinstance(x, (str, typing.Quantity)):
    nondim = sim_units.nondimensionalize(typing.Quantity(x))
    assert isinstance(nondim, typing.Numeric)  # make pytype happy.
    return nondim
  return x


def get_default_units() -> SimUnits:
  """units.DEFAULT_UNITS fn-wrapper for fiddle configs."""
  return DEFAULT_UNITS


def get_one_minute_dt_units() -> SimUnits:
  """units.DEFAULT_UNITS_RESCALED fn-wrapper for fiddle configs."""
  return ONE_MINUTE_DT_UNITS


def nondimensionalize_timedelta64(
    timedelta64: np.timedelta64, sim_units: SimUnits
) -> typing.Numeric:
  """sim_units.nondimensionalize_timedelta64 fn-wrapper for fiddle configs."""
  return sim_units.nondimensionalize_timedelta64(timedelta64)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from absl.testing import absltest
import jax
import neuralgcm
import numpy as np
import xarray


def _assert_allclose(actual, desired, *, err_msg=None, range_rtol=1e-5):
  span = desired.max() - desired.min()
  np.testing.assert_allclose(
      actual / span, desired / span, rtol=0, atol=range_rtol, err_msg=err_msg
  )


def load_tl63_stochastic_model():
  return neuralgcm.PressureLevelModel.from_checkpoint(
      neuralgcm.demo.load_checkpoint_tl63_stochastic()
  )


class APITest(absltest.TestCase):

  def assertDictEqual(self, actual, desired, *, err_msg=None):
    self.assertEqual(sorted(actual.keys()), sorted(desired.keys()))
    for key in actual:
      x = actual[key]
      y = desired[key]
      err_msg2 = key if err_msg is None else f'{err_msg}/{key}'
      if isinstance(y, dict):
        self.assertDictAllclose(x, y, err_msg=err_msg2)
      else:
        np.testing.assert_array_equal(x, y, err_msg=err_msg2)

  def assertDictAllclose(
      self, actual, desired, *, err_msg=None, range_rtol=1e-5
  ):
    self.assertEqual(sorted(actual.keys()), sorted(desired.keys()))
    for key in actual:
      x = actual[key]
      y = desired[key]
      err_msg2 = key if err_msg is None else f'{err_msg}/{key}'
      if isinstance(y, dict):
        self.assertDictAllclose(x, y, err_msg=err_msg2, range_rtol=range_rtol)
      else:
        _assert_allclose(x, y, err_msg=err_msg2, range_rtol=range_rtol)

  def test_model_properties(self):
    model = load_tl63_stochastic_model()

    self.assertEqual(model.timestep, np.timedelta64(1, 'h'))

    expected_inputs = [
        'geopotential',
        'specific_humidity',
        'temperature',
        'u_component_of_wind',
        'v_component_of_wind',
        'specific_cloud_ice_water_content',
        'specific_cloud_liquid_water_content',
    ]
    self.assertEqual(model.input_variables, expected_inputs)

    expected_forcings = [
        'sea_ice_cover',
        'sea_surface_temperature',
    ]
    self.assertEqual(model.forcing_variables, expected_forcings)

    self.assertEqual(model.data_coords.nodal_shape, (37, 128, 64))

    self.assertEqual(model.model_coords.nodal_shape, (32, 128, 64))

  def test_to_and_from_nondim_units(self):
    model = load_tl63_stochastic_model()

    nondim = model.to_nondim_units(1.0, 'meters')
    self.assertAlmostEqual(nondim, 1 / 6.37122e6)  # Earth radius units

    roundtripped = model.from_nondim_units(nondim, 'meters')
    self.assertAlmostEqual(roundtripped, 1.0)

  def test_sim_time_utilities(self):
    model = load_tl63_stochastic_model()
    origin = model.datetime64_to_sim_time(np.datetime64('1979-01-01T00:00:00'))
    self.assertEqual(origin, 0.0)

    original = np.datetime64('1959-01-01T00:00:00')
    roundtripped = model.sim_time_to_datetime64(
        model.datetime64_to_sim_time(original)
    )
    self.assertEqual(original, roundtripped)

  def test_from_xarray(self):
    model = load_tl63_stochastic_model()
    ds = neuralgcm.demo.load_data(model.data_coords)

    expected_inputs = {k: ds[k].values for k in model.input_variables}
    expected_inputs['sim_time'] = np.array([-92034.607104])

    expected_forcings = {
        k: ds[k].values[:, np.newaxis, :, :] for k in model.forcing_variables
    }
    expected_forcings['sim_time'] = np.array([-92034.607104])

    inputs = model.inputs_from_xarray(ds)
    self.assertDictEqual(inputs, expected_inputs)

    forcings = model.forcings_from_xarray(ds)
    self.assertDictEqual(forcings, expected_forcings)

    inputs, forcings = model.data_from_xarray(ds)
    self.assertDictEqual(inputs, expected_inputs)
    self.assertDictEqual(forcings, expected_forcings)

  def test_stochastic_model_basics(self):
    timesteps = 3
    dt = np.timedelta64(1, 'h')

    model = load_tl63_stochastic_model()
    ds_in = neuralgcm.demo.load_data(model.data_coords)

    data_in, forcings_in = model.data_from_xarray(ds_in.isel(time=0))
    persistence_forcings = model.forcings_from_xarray(ds_in.head(time=1))

    # run model
    encoded = model.encode(data_in, forcings_in, rng_key=jax.random.key(0))
    _, data_out = model.unroll(
        encoded,
        persistence_forcings,
        steps=timesteps,
        timedelta=dt,
        start_with_input=True,
    )

    # convert to xarray
    t0 = ds_in.time.values[0]
    times = np.arange(t0, t0 + timesteps * dt, dt)
    ds_out = model.data_to_xarray(data_out, times=times)

    # validate decoded data
    actual = ds_out.head(time=1)

    sim_time = model.datetime64_to_sim_time(ds_in.time.data)
    expected = ds_in.drop_vars(
        ['sea_surface_temperature', 'sea_ice_cover']
    ).assign(sim_time=('time', sim_time))

    # check matching variable shapes
    xarray.testing.assert_allclose(actual, expected, atol=1e6)

    # check that round-tripping the initial condition is approximately correct
    typical_relative_error = abs(actual - expected).median() / expected.std()
    tolerance = xarray.Dataset({
        'u_component_of_wind': 0.04,
        'v_component_of_wind': 0.08,
        'temperature': 0.02,
        'geopotential': 0.0005,
        'specific_humidity': 0.003,
        'specific_cloud_liquid_water_content': 0.12,
        'specific_cloud_ice_water_content': 0.15,
    })
    self.assertTrue(
        (typical_relative_error < tolerance).to_array().values.all(),
        msg=f'typical relative error is too large:\n{typical_relative_error}',
    )

    # test decode()
    decoded = model.decode(encoded, forcings_in)
    expected = jax.tree.map(lambda x: x[0, ...], data_out)
    self.assertDictAllclose(decoded, expected)

    # test advance()
    self.assertEqual(model.timestep, dt)
    decoded2 = model.decode(model.advance(encoded, forcings_in), forcings_in)
    expected2 = jax.tree.map(lambda x: x[1, ...], data_out)
    self.assertDictAllclose(decoded2, expected2)

    # TODO(shoyer): verify RNG key works correctly
    # TODO(shoyer): verify RNG key is optional for deterministic models

  def test_encoded_state(self):
    model = load_tl63_stochastic_model()
    ds_in = neuralgcm.demo.load_data(model.data_coords)
    data_in, forcings_in = model.data_from_xarray(ds_in.isel(time=0))

    encoded = model.encode(data_in, forcings_in, rng_key=jax.random.key(0))

    ds_advanced = model.data_to_xarray(
        {
            'vorticity': encoded.state.vorticity,
            'divergence': encoded.state.divergence,
        },
        times=None,
        decoded=False,
    )

    dims = ('level', 'longitudinal_mode', 'total_wavenumber')
    z, x, y = model.model_coords.modal_shape
    levels = np.linspace(1 - 1/32, 1 / 32, 32)
    longitudinal_modes = np.array([(i//2) * (-1)**i for i in range(128)])
    total_wavenumbers = np.arange(65)
    expected = xarray.Dataset(
        {
            'vorticity': (dims, np.zeros((z, x, y))),
            'divergence': (dims, np.zeros((z, x, y))),
        },
        coords={
            'level': levels,
            'longitudinal_mode': longitudinal_modes,
            'total_wavenumber': total_wavenumbers,
        },
    )
    # verify expected shpaes
    xarray.testing.assert_allclose(ds_advanced, expected, atol=1e6)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines `diagnostic` modules that compute diagnostic predictions."""

from collections import abc
from typing import Any, Callable, Optional, Protocol

from dinosaur import coordinate_systems
from dinosaur import scales
from dinosaur import sigma_coordinates
from dinosaur import typing

import gin
import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np


TransformModule = typing.TransformModule

PRECIPITATION = 'precipitation'
EVAPORATION = 'evaporation'


class DiagnosticFn(Protocol):
  """Implements initialization and computation of model diagnostic fields."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: dict[str, Any],
  ):
    del coords, dt, physics_specs, aux_features

  def __call__(
      self,
      model_state: typing.ModelState,
      physics_tendencies: typing.Pytree,
      forcing: typing.Forcing | None = None,
  ) -> dict[str, jax.Array]:
    """Computes diagnostic field from `model_state` and `physics_tendencies`."""
    ...


DiagnosticModule = Callable[..., DiagnosticFn]


@gin.register
class NoDiagnostics:
  """Diagnostic module that computes no diagnostics."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: dict[str, Any],
  ):
    del coords, dt, physics_specs, aux_features

  def __call__(
      self,
      model_state: typing.ModelState,
      physics_tendencies: typing.Pytree,
      forcing: typing.Forcing | None = None,
  ) -> dict[str, jax.Array]:
    return {}


@gin.register
class CombinedDiagnostics:
  """Computes a combination of multiple diagnostics."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: dict[str, Any],
      diagnostic_modules: abc.Sequence[DiagnosticModule] = gin.REQUIRED,
  ):
    self.diagnostic_fns = [
        module(coords, dt, physics_specs, aux_features)
        for module in diagnostic_modules
    ]

  def __call__(
      self,
      model_state: typing.ModelState,
      physics_tendencies: typing.Pytree,
      forcing: typing.Forcing | None = None,
  ) -> dict[str, jax.Array]:
    diagnostics = {}
    for fn in self.diagnostic_fns:
      new_diagnostics = fn(model_state, physics_tendencies, forcing)
      if any(k in diagnostics for k in new_diagnostics):
        raise ValueError(
            f'{new_diagnostics.keys()} overlaps with {diagnostics.keys()}'
        )
      diagnostics.update(new_diagnostics)
    return diagnostics


@gin.register
class PrecipitationMinusEvaporationDiagnostics:
  """Computes `P-E` by integrating physics_tendencies.

  Depending on the `method` computes either precipitation minus evaporation
  rate, which in ERA5 has units `kg m**-2 s**-1` or time-accumulated value
  in `kg m**-2` if `method == cumulative`.
  """

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: dict[str, Any],
      moisture_species: tuple[str, ...] = (
          'specific_humidity',
          'specific_cloud_ice_water_content',
          'specific_cloud_liquid_water_content',
      ),
      method: str = 'rate',
  ):
    del aux_features
    self.coords = coords
    self.dt = dt
    self.physics_specs = physics_specs
    self.moisture_species = moisture_species
    self.method = method
    self.to_nodal_fn = coords.horizontal.to_nodal

  def _compute_evaporation_minus_precipitation(
      self, model_state: typing.ModelState, physics_tendencies: typing.Pytree
  ) -> typing.Array:
    """Computes evaporation minus precipitation."""
    lsp = model_state.state.log_surface_pressure
    p_surface = jnp.squeeze(jnp.exp(self.to_nodal_fn(lsp)), axis=0)
    moisture_tendencies = [
        v
        for tracer, v in physics_tendencies.tracers.items()
        if tracer in self.moisture_species
    ]
    moisture_tendencies = sum(self.to_nodal_fn(moisture_tendencies))
    scale = p_surface / self.physics_specs.g
    e_minus_p = scale * sigma_coordinates.sigma_integral(
        moisture_tendencies, self.coords.vertical, keepdims=False
    )
    return e_minus_p

  def __call__(
      self,
      model_state: typing.ModelState,
      physics_tendencies: typing.Pytree,
      forcing: typing.Forcing | None = None,
  ) -> typing.Pytree:
    """Computes precipitation minus evaporation."""
    del forcing  # unused
    e_minus_p = self._compute_evaporation_minus_precipitation(
        model_state, physics_tendencies
    )
    if self.method == 'rate':
      return {'P_minus_E_rate': -e_minus_p}
    elif self.method == 'cumulative':
      # TODO(dkochkov) Address possible precision loss due to small deltas.
      surface_nodal_shape = self.coords.horizontal.nodal_shape
      previous = model_state.diagnostics.get(
          'P_minus_E_cumulative',
          jnp.zeros(surface_nodal_shape))
      return {'P_minus_E_cumulative': previous - (e_minus_p * self.dt)}
    else:
      raise ValueError(f'Unknown {self.method=}, must be `rate`/`cumulative`')


@gin.register
class PrecipitableWaterDiagnostics:
  """Computes cumulative preciptable water in the state."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: dict[str, Any],
      moisture_species: tuple[str, ...] = (
          'specific_humidity',
          'specific_cloud_ice_water_content',
          'specific_cloud_liquid_water_content',
      ),
  ):
    del dt, aux_features
    self.coords = coords
    self.physics_specs = physics_specs
    self.moisture_species = moisture_species
    self.to_nodal_fn = coords.horizontal.to_nodal

  def __call__(
      self,
      model_state: typing.ModelState,
      physics_tendencies: typing.Pytree,
      forcing: typing.Forcing | None = None,
  ) -> typing.Pytree:
    """Computes preciptable water."""
    del physics_tendencies, forcing  # unused
    lsp = model_state.state.log_surface_pressure
    p_surface = jnp.squeeze(jnp.exp(self.to_nodal_fn(lsp)), axis=0)
    moisture_tracers = [
        v
        for tracer, v in model_state.tracers.items()
        if tracer in self.moisture_species
    ]
    moisture = sum(self.to_nodal_fn(moisture_tracers))
    water_density = self.physics_specs.nondimensionalize(scales.WATER_DENSITY)
    scale = p_surface / (self.physics_specs.g * water_density)
    water = scale * sigma_coordinates.sigma_integral(
        moisture, self.coords.vertical, keepdims=False
    )
    return {'precipitable_water': water}


@gin.register
class NodalModelDiagnosticsDecoder:
  """Diagnostics decoder that returns elements from model_state.diagnostics."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: dict[str, Any],
  ):
    del dt, aux_features
    self.coords = coords
    self.physics_specs = physics_specs

  def __call__(
      self,
      model_state: typing.ModelState,
      physics_tendencies: typing.Pytree,
      forcing: typing.Forcing | None = None,
  ) -> typing.Pytree:
    """Computes precipitation minus evaporation."""
    del physics_tendencies, forcing  # unused.
    nodal_diagnostics = coordinate_systems.maybe_to_nodal(
        model_state.diagnostics, self.coords
    )
    return nodal_diagnostics


# TODO(janniyuval) add a decoder that can add some Gaussian noise to evap/precip
@gin.register
class PrecipitationDiagnosticsConstrained(
    hk.Module, PrecipitationMinusEvaporationDiagnostics
):
  """Predict evaporation and computes cumulative precipitation.

  Calculation is based on calculating `P-E` by integrating physics_tendencies.
  Depending on the `method` computes either precipitation
  rate, (which in ERA5 has units `kg m**-2 s**-1`) or time-accumulated value
  in `Length` units (GPCP uses mm/day) if `method == cumulative`.
  Evaporation has the units of `kg m**-2 s**-1` in ERA5.
  """

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: dict[str, Any],
      embedding_module: typing.EmbeddingModule,
      moisture_species: tuple[str, ...] = (
          'specific_humidity',
          'specific_cloud_ice_water_content',
          'specific_cloud_liquid_water_content',
      ),
      is_precipitation: bool = True,
      method_precipitation: str = 'cumulative',
      method_evaporation: str = 'rate',
      name: Optional[str] = None,
      field_name: str = 'total_precipitation',
  ):
    # del aux_features
    super().__init__(name=name)
    self.coords = coords
    self.dt = dt
    self.physics_specs = physics_specs
    self.moisture_species = moisture_species
    self.method_precipitation = method_precipitation
    self.method_evaporation = method_evaporation
    self.to_nodal_fn = coords.horizontal.to_nodal
    self.is_precipitation = is_precipitation
    if self.is_precipitation:
      predicted_name = PRECIPITATION
      diagnosed_name = EVAPORATION
    else:
      predicted_name = EVAPORATION
      diagnosed_name = PRECIPITATION

    self.predicted_name = predicted_name
    self.diagnosed_name = diagnosed_name

    output_shapes = {
        f'{predicted_name}': np.asarray(coords.surface_nodal_shape)
    }

    self.embedding_fn = embedding_module(
        coords, dt, physics_specs, aux_features, output_shapes=output_shapes
    )
    self.water_density = self.physics_specs.nondimensionalize(
        scales.WATER_DENSITY
    )
    self.field_name = field_name

  def __call__(
      self,
      model_state: typing.ModelState,
      physics_tendencies: typing.Pytree,
      forcing: typing.Forcing | None = None,
  ) -> typing.Pytree:
    """Computes precipitation minus evaporation."""
    e_minus_p = self._compute_evaporation_minus_precipitation(
        model_state, physics_tendencies
    )
    water_budget = self.embedding_fn(
        model_state.state,
        model_state.memory,
        model_state.diagnostics,
        model_state.randomness,
        forcing,
    )
    water_budget[self.diagnosed_name] = (
        -e_minus_p - water_budget[self.predicted_name]
    )

    # Note: In ERA5 mean_evaporation_rate (kg m**-2 s**-1)
    # is negative for evaporation.
    # In GPCP precipitation is positive (mm/day).
    # Here e_minus_p is positive for evaporation.
    output_dict = {}
    surface_nodal_shape = self.coords.horizontal.nodal_shape
    if self.method_precipitation == 'rate':  # units: length/time
      output_dict[PRECIPITATION + '_rate'] = (
          water_budget[PRECIPITATION]
      ) / self.water_density
    elif self.method_precipitation == 'cumulative':  # units: length
      previous = model_state.diagnostics.get(
          self.field_name, jnp.zeros(surface_nodal_shape)
      )
      # TODO(janniyuval) remove precipitation_cumulative_mean once no models
      # use it.
      assert self.field_name in [
          'total_precipitation',
          'precipitation_cumulative_mean',
      ], self.field_name
      output_dict[self.field_name] = previous + (
          (water_budget[PRECIPITATION] / self.water_density) * self.dt
      )
    else:
      raise ValueError(
          f'Precipitation method is {self.method_precipitation=}, but it must'
          ' be `rate`/`cumulative`'
      )
    if self.method_evaporation == 'rate':  # units: mass length**-2 time**-1
      output_dict[EVAPORATION] = water_budget[EVAPORATION]
    elif self.method_evaporation == 'cumulative':  # units: length
      previous_evap = model_state.diagnostics.get(
          EVAPORATION + '_cumulative', jnp.zeros(surface_nodal_shape)
      )
      output_dict[EVAPORATION + '_cumulative'] = (
          previous_evap
          + (water_budget[EVAPORATION] / self.water_density) * self.dt
      )
    else:
      raise ValueError(
          f'Evaporation method is {self.method_evaporation=},  but it must be'
          ' `rate`/`cumulative`'
      )
    return output_dict


@gin.register
class SurfacePressureDiagnostics:
  """Getting the surface pressure of the state."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: dict[str, Any],
  ):
    del dt, aux_features, physics_specs
    self.to_nodal_fn = coords.horizontal.to_nodal

  def __call__(
      self,
      model_state: typing.ModelState,
      physics_tendencies: typing.Pytree,
      forcing: typing.Forcing | None = None,
  ) -> typing.Pytree:
    """Computes surface pressure."""
    del physics_tendencies, forcing  # unused
    lsp = model_state.state.log_surface_pressure
    surface_pressure = jnp.squeeze(jnp.exp(self.to_nodal_fn(lsp)), axis=0)
    return {'surface_pressure': surface_pressure}



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Training utility functions for NeuralGCM."""

import collections
from collections import abc
import functools
import logging
import math
from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    Mapping,
    Optional,
    Tuple,
    TypeVar,
    Union,
)
from dinosaur import pytree_utils
from dinosaur import typing

import einops
import gin
import haiku as hk
import jax
from jax.experimental import mesh_utils
import jax.numpy as jnp
from neuralgcm import optimization
import numpy as np
import optax


# pylint: disable=logging-fstring-interpolation


PRNGKeyArray = typing.PRNGKeyArray
Array = Union[np.ndarray, jnp.ndarray]
PyTree = Any
Forcing = typing.Forcing

IntOrArray = Union[int, Array]
OptState = optimization.OptState
ModelParams = Any
ModelGradients = ModelParams
EMAParams = ModelParams
StepAndOptState = Tuple[IntOrArray, OptState]
StepOptAndEMAState = Tuple[IntOrArray, OptState, ModelParams]
LossValue = Array
LossFunction = Callable[[PyTree, PyTree], LossValue]
LossAndGradFunction = Callable[
    [ModelParams, PRNGKeyArray, PyTree, Forcing],
    Tuple[LossValue, ModelGradients],
]
MetricFunction = Callable[[PyTree, PyTree], Union[Array, Mapping[str, Array]]]
TrainStepFunction = Callable[
    [PRNGKeyArray, StepAndOptState, PyTree, Forcing],
    Tuple[StepAndOptState, LossValue],
]
EvalStepFunction = Callable[
    [ModelParams, PRNGKeyArray, PyTree, Forcing], Mapping[str, Array]
]
TrajectoryFunction = Callable[
    [ModelParams, PRNGKeyArray, PyTree, Forcing], Tuple[PyTree, PyTree]
]


def flatten_dict(
    inputs: Mapping[str, Any],
    parent_key: str = '',
    sep: str = ' ',
) -> Mapping[str, Array]:
  """Returns a flattened version of `inputs` dictionary."""
  items = []
  for k, v in inputs.items():
    new_key = parent_key + sep + k if parent_key else k
    if isinstance(v, Mapping):
      items.extend(flatten_dict(v, new_key, sep=sep).items())
    else:
      items.append((new_key, v))
  keys, counts = np.unique(np.array([x[0] for x in items]), return_counts=True)
  if (counts > 1).any():
    raise ValueError(f'got duplicate keys {keys[counts > 1]}')
  return dict(items)


#
#  Note that all functions below deal with *batched* inputs.
#


def loss_and_gradient(
    trajectory_fn: TrajectoryFunction,
    loss_fn: LossFunction,
) -> LossAndGradFunction:
  """Returns a function that computes loss and the gradient of the loss.

  Args:
    trajectory_fn: a function that accepts `params` and `initial_velocity` and
      returns a trajectory of velocities.
    loss_fn: a function that accepts a predicted trajectory and a ground truth
      trajectory, returning a scalar loss value.

  Returns:
    A function that accepts `params, initial_velocity, target_trajectory` and
    returns the loss and the gradient of the loss.
  """

  def _loss(
      params: ModelParams,
      rng: PRNGKeyArray,
      target_trajectory: PyTree,
      forcing_data: typing.ForcingData,
  ) -> LossValue:
    """Returns loss value and gradient with respect to model parameters."""
    _, predicted_trajectory = trajectory_fn(
        params, rng, target_trajectory, forcing_data
    )
    loss = loss_fn(predicted_trajectory, target_trajectory)  # type: ignore
    return loss

  return jax.value_and_grad(_loss)


def train_step(
    loss_and_grad_fn: LossAndGradFunction,
    optimizer: optax.GradientTransformation,
) -> TrainStepFunction:
  """Returns a function that performs a single training step.

  Args:
    loss_and_grad_fn: a function that accepts `params, initial_velocity,
      target_trajectory` and returns the loss and the gradient of the loss.
    optimizer: Optax optimizer to update params and internal state.

  Returns:
    A function that performs a single training step.
  """

  def _train_step(
      rng: PRNGKeyArray,
      step_and_state: StepAndOptState,
      target_trajectory: PyTree,
      forcing_data: typing.ForcingData,
  ) -> Tuple[StepAndOptState, LossValue]:
    """A function that performs a single training step."""
    step, opt_state = step_and_state
    loss, grad = loss_and_grad_fn(
        opt_state.params, rng, target_trajectory, forcing_data
    )

    updates, new_state = optimizer.update(
        grad, opt_state.state, opt_state.params
    )
    new_params = optax.apply_updates(opt_state.params, updates)
    new_opt_state = OptState(state=new_state, params=new_params)

    return (step + 1, new_opt_state), loss

  return _train_step


def eval_batch(
    trajectory_fn: TrajectoryFunction,
    metric_funcs: Mapping[str, MetricFunction],
) -> EvalStepFunction:
  """Returns a function that performs a single evaluation step.

  Args:
    trajectory_fn: a function that accepts `params` and `initial_velocity` and
      returns a trajectory of velocities.
    metric_funcs: a dictionary mapping strings to metric funcutils, each
      returning either a metric scalar or a dictionary of such.

  Returns:
    A function that performs a single evaluation step.
  """

  def _eval_batch(
      params: ModelParams,
      rng: PRNGKeyArray,
      target_trajectory: PyTree,
      forcing_data: typing.ForcingData,
  ) -> Mapping[str, Array]:
    """A function that performs a single evaluation step."""
    _, predicted_trajectory = trajectory_fn(
        params, rng, target_trajectory, forcing_data
    )
    metric_values = {
        k: metric(predicted_trajectory, target_trajectory)
        for k, metric in metric_funcs.items()
    }
    results = flatten_dict(metric_values)
    return results

  return _eval_batch


def streaming_mean(
    rngs: Iterable[PRNGKeyArray],
    batch_and_forcing: Iterable[Tuple[PyTree, Forcing]],
    eval_fn: Callable[[PRNGKeyArray, PyTree, Forcing], Mapping[str, Array]],
    data_preprocess_fn: Callable[..., PyTree] = lambda x: x,
) -> Mapping[str, Array]:
  """Runs evaluation on `eval_data`.

  Args:
    rngs: an iterable of random number keys to be used for evaluation.
    batch_and_forcing: an iterable of batched velocity trajectories and forcing.
    eval_fn: a function that performs a single evaluation step.
    data_preprocess_fn: a preprocessing function be applied to each batch.

  Returns:
    A dict mapping strings to metric values.

  Raises:
    RuntimeError: if there are no batches to iterate over.
  """
  eval_metrics = collections.defaultdict(float)
  count = 0
  for rng, (batch, forcing) in zip(rngs, batch_and_forcing):
    batch = data_preprocess_fn(batch)
    batch_metrics = eval_fn(rng, batch, forcing)
    for k, v in batch_metrics.items():
      eval_metrics[k] += v
    count += 1
  if not count:
    raise RuntimeError('no batches to iterate over')
  return {k: v / count for k, v in eval_metrics.items()}


@gin.register
def identity(batch: Tuple[Array, ...], rng: Array = None) -> Tuple[Array, ...]:  # pytype: disable=annotation-type-mismatch  # jax-ndarray
  """Identity preprocessing function that does not modify the `batch`."""
  del rng  # unused.
  return batch


@gin.configurable
def add_noise_to_input_frame(
    batch: Tuple[Array, ...], rng: Array, scale: float = 1e-2, **kwargs
) -> Tuple[Array, ...]:
  """Adds noise to the 0th time frame in the `batch`.

  Args:
    batch: original batch to which the noise will be added.
    rng: random number key to be used to generate noise.
    scale: scale of the normal noise to be added.
    **kwargs: other keyword arguments. Not used.

  Returns:
    batch with noise added along the 0th time slice.
  """
  del kwargs  # unused.
  time_zero_slice = pytree_utils.slice_along_axis(batch, 1, 0)
  shapes = jax.tree.map(np.shape, time_zero_slice)
  rngs = jax.random.split(rng, len(jax.tree.leaves(time_zero_slice)))
  rngs = jax.tree.unflatten(jax.tree.structure(time_zero_slice), rngs)

  def noise_fn(key, s):
    return scale * jax.random.truncated_normal(key, -2.0, 2.0, s)

  noise = jax.tree.map(noise_fn, rngs, shapes)
  add_noise_fn = lambda x, n: x.at[:, 0, ...].add(n)
  return jax.tree.map(add_noise_fn, batch, noise)


def preprocess(
    data_iterator: Iterator[Tuple[Array, ...]],
    rng_stream: Iterator[Array],
    preprocess_fn: Callable[..., Tuple[Array, ...]],
):
  """Generator that applies `preprocess_fn` to entries of the `data_iterator`.

  Args:
    data_iterator: numpy iterator holding the data.
    rng_stream: stream of random numbers to be used by `preprocess_fn`.
    preprocess_fn: preprocessing function to be applied to each batch of data.

  Yields:
    Batch of data from `data_iterator` preprocessed with `preprocess_fn`.
  """
  preprocess_fn = jax.jit(preprocess_fn)
  while True:
    rng = next(rng_stream)
    yield preprocess_fn(next(data_iterator), rng)


def split_rngs(rngs: PRNGKeyArray, num: int) -> PRNGKeyArray:
  """Splits `rngs` into `num` along the last batch axis."""
  ndim = rngs.ndim
  split_fn = jax.random.split
  for _ in range(ndim - 1):
    split_fn = jax.vmap(split_fn, (0, None), 1)
  return split_fn(rngs, num)


@functools.partial(jax.jit, static_argnames=['batch_shape'])
def _split_rmgs_by_batch_shape(
    rngs: PRNGKeyArray,
    batch_shape: tuple[int, ...],
) -> PRNGKeyArray:
  for batch_size in batch_shape[::-1]:
    rngs = split_rngs(rngs, batch_size)
  return rngs


class BatchedPRNGSequence(Iterator):
  """Iterator of JAX different random keys split by `batch_shape`."""

  def __init__(
      self,
      key_or_seed: Union[int, PRNGKeyArray],
      batch_shape: Optional[Tuple[int, ...]] = None,
  ):
    """Creates an instance a class.

    Args:
      key_or_seed: Key or seed to initialize the random sequence.
      batch_shape: Batch shape of the sequence.
    """
    self._key = hk.PRNGSequence(key_or_seed)
    self.batch_shape = batch_shape

  def reserve(self, num: int):
    """Splits an additional ``num`` keys for later use."""
    self._key = self._key.reserve(num)

  def __next__(self):
    rngs = next(self._key)
    return _split_rmgs_by_batch_shape(rngs, self.batch_shape)


@jax.jit
def _combine_rng_seeds(seeds: jax.Array) -> jax.Array:
  key = jax.random.PRNGKey(seeds[0])
  for seed in seeds[1:]:
    key = jax.random.fold_in(key, seed)
  return jax.random.bits(key, shape=(), dtype=jnp.uint32)


def combine_rng_seeds(*seeds: int) -> int:
  """Combine uint32 seeds into a single Python integer RNG seed."""
  # Put the seeds on the first CPU device so that JAX runs the entire
  # computation on the CPU.
  seeds = jax.device_put(
      np.array(seeds), device=jax.local_devices(backend='cpu')[0]
  )
  return int(_combine_rng_seeds(seeds))


def ensure_sharded_rng_key(
    rng_key: jax.Array, *, mesh: jax.sharding.Mesh
) -> jax.Array:
  """Ensure that a batched PRNG key is sharded across all devices."""
  spec = P('batch', 'ensemble', None)
  sharding = jax.sharding.NamedSharding(mesh, spec)
  return jax.lax.with_sharding_constraint(rng_key, sharding)


def get_tpu_physical_mesh_shape() -> tuple[int, int, int] | None:
  """Get the shape of the TPU connectivity torus for v4 or v5 chips."""
  jax_devices = jax.devices()
  try:
    device_coords = [d.coords for d in jax_devices]
  except AttributeError:
    return None  # no "coords" attribute (e.g., using CPU devices)
  dims = tuple(d + 1 for d in max(device_coords))
  if len(dims) != 3 or math.prod(dims) != len(jax_devices):
    return None
  return dims


# dict of dicts of indicating how to rearrange from physical TPU mesh layouts
# (X, Y, Z) into logical mesh layouts (batch, ensemble, z, x, y) with
# einops.rearrange for model training.
# {tpu_topology: {(ensemble_shards, z_shard, x_shards, y_shards): ...}}
_TPU_LAYOUT_REARRANGEMENTS = {
    '2x2x2': {
        (1, 1, 1, 1): 'b0 b1 b2 -> (b0 b1 b2) () () () ()',
        (1, 2, 1, 1): 'z b0 b1 -> (b0 b1) () z () ()',
        (2, 1, 1, 1): 'e b0 b1 -> (b0 b1) e () () ()',
    },
    '2x2x4': {
        (1, 1, 1, 1): 'b0 b1 b2 -> (b0 b1 b2) () () () ()',
        (1, 2, 1, 1): 'z b0 b1 -> (b0 b1) () z () ()',
        (1, 4, 1, 1): 'b0 b1 z -> (b0 b1) () z () ()',
        (2, 1, 1, 1): 'e b0 b1 -> (b0 b1) e () () ()',
    },
    '2x4x4': {
        (1, 1, 1, 1): 'b0 b1 b2 -> (b0 b1 b2) () () () ()',
        (1, 2, 1, 1): 'z b0 b1 -> (b0 b1) () z () ()',
        (1, 4, 1, 1): 'b0 b1 z -> (b0 b1) () z () ()',
        (2, 1, 1, 1): 'e b0 b1 -> (b0 b1) e () () ()',
        (2, 2, 1, 1): 'z (b0 e) b1 -> (b0 b1) e z () ()',
    },
    '4x4x4': {
        (1, 1, 1, 1): 'b0 b1 b2 -> (b0 b1 b2) () () () ()',
        (1, 2, 1, 1): '(b0 z) b1 b2 -> (b0 b1 b2) () z () ()',
        (1, 4, 1, 1): 'b0 b1 z -> (b0 b1) () z () ()',
        (1, 2, 2, 1): '(b0 z) (b1 x) b2 -> (b0 b1 b2) () z x ()',
        (2, 1, 1, 1): '(b0 e) b1 b2 -> (b0 b1 b2) e () () ()',
        (2, 2, 1, 1): '(b0 e) (b1 z) b2 -> (b0 b1 b2) e z () ()',
        (2, 2, 2, 1): '(b0 e) (b1 z) (b2 x) -> (b0 b1 b2) e z x ()',
    },
    '4x4x8': {
        (1, 1, 1, 1): 'b0 b1 b2 -> (b0 b1 b2) () () () ()',
        (1, 4, 2, 1): 'z (b0 x) b1 -> (b0 b1) () z x ()',
        (1, 4, 2, 2): 'z (b0 x) (b1 y) -> (b0 b1) () z x y',
        (2, 4, 2, 1): 'z (b0 x) (b1 e) -> (b0 b1) e z x ()',
    },
    '4x8x8': {
        (1, 1, 1, 1): 'b0 b1 b2 -> (b0 b1 b2) () () () ()',
        (1, 4, 2, 1): 'z (b0 x) b1 -> (b0 b1) () z x ()',
        (1, 4, 2, 2): 'z (b0 x) (b1 y) -> (b0 b1) () z x y',
        (2, 4, 2, 1): 'z (b0 e) (b1 x) -> (b0 b1) e z x ()',
        (2, 4, 2, 2): 'z (b0 e x) (b1 y) -> (b0 b1) e z x y',
    },
    '2x2x1': {
        (1, 1, 1, 1): 'b0 b1 () -> (b0 b1) () () () ()',
        (1, 2, 1, 1): 'z b0 () -> b0 () z () ()',
        (2, 1, 1, 1): 'e b0 () -> b0 e () () ()',
        (2, 2, 1, 1): 'e z () -> () e z () ()',
    },
    '2x4x1': {
        (1, 1, 1, 1): 'b0 b1 () -> (b0 b1) () () () ()',
        (1, 2, 1, 1): 'z b0 () -> b0 () z () ()',
        (2, 1, 1, 1): 'e b0 () -> b0 e () () ()',
        (2, 2, 1, 1): 'z (b0 e) -> b0 e z () ()',
    },
    '4x4x1': {
        (1, 1, 1, 1): 'b0 b1 () -> (b0 b1) () () () ()',
        (1, 2, 1, 1): '(b0 z) b1 () -> (b0 b1) () z () ()',
        (2, 1, 1, 1): '(b0 e) b1 () -> (b0 b1) e () () ()',
        (2, 2, 1, 1): '(b0 e) (b1 z) () -> (b0 b1) e z () ()',
    },
    '4x8x1': {
        (1, 1, 1, 1): 'b0 b1 () -> (b0 b1) () () () ()',
        (1, 2, 1, 1): '(b0 z) b1 () -> (b0 b1) () z () ()',
        (1, 2, 2, 1): '(b0 z) (b1 x) () -> (b0 b1) () z x ()',
        (2, 1, 1, 1): '(b0 e) b1 () -> (b0 b1) e () () ()',
        (2, 2, 1, 1): '(b0 e) (b1 z) () -> (b0 b1) e z () ()',
    },
    '8x8x1': {
        (1, 1, 1, 1): 'b0 b1 () -> (b0 b1) () () () ()',
        (1, 2, 1, 1): '(b0 z) b1 () -> (b0 b1) () z () ()',
        (1, 2, 2, 1): '(b0 z) (b1 x) () -> (b0 b1) () z x ()',
        (2, 1, 1, 1): '(b0 e) b1 () -> (b0 b1) e () () ()',
        (2, 2, 1, 1): '(b0 e) (b1 z) () -> (b0 b1) e z () ()',
        (2, 2, 2, 1): '(b0 e z) (b1 x) () -> (b0 b1) e z x ()',
    },
    '8x16x1': {
        (1, 1, 1, 1): 'b0 b1 () -> (b0 b1) () () () ()',
        (1, 2, 1, 1): '(b0 z) b1 () -> (b0 b1) () z () ()',
        (1, 2, 2, 1): '(b0 z) (b1 x) () -> (b0 b1) () z x ()',
        (2, 1, 1, 1): '(b0 e) b1 () -> (b0 b1) e () () ()',
        (2, 2, 1, 1): '(b0 e) (b1 z) () -> (b0 b1) e z () ()',
        (2, 2, 2, 1): '(b0 e z) (b1 x) () -> (b0 b1) e z x ()',
    },
}


def create_spmd_mesh(sizes: dict[str, int]) -> jax.sharding.Mesh:
  """Create an SPMD mesh suitable for data & model parallelism.

  Args:
    sizes: dictionary mapping from dimension names (batch, z, x, and y) to the
      number of devices desired along that axis in the parallel mesh.

  Returns:
    Mesh with axis names ['batch', 'ensemble', 'x', 'y', 'z'] and the desired
    axis sizes.
  """
  axis_names = ['batch', 'ensemble', 'z', 'x', 'y']
  for name in sizes:
    if name not in axis_names:
      raise ValueError(f'unrecognized {name!r} not in {axis_names}')

  logical_mesh_shape = tuple(
      sizes.get(axis_name, 1) for axis_name in axis_names
  )
  if math.prod(logical_mesh_shape) != jax.device_count():
    raise ValueError(
        f'{logical_mesh_shape=} is incompatible with {jax.device_count()=}'
    )

  physical_mesh_shape = get_tpu_physical_mesh_shape()
  if physical_mesh_shape is None:
    try:
      # only succeeds if the logical mesh shape perfectly matches the physical
      # mesh, e.g., in the case of pure data parallelism
      mesh_devices = mesh_utils.create_device_mesh(logical_mesh_shape)
    except (AssertionError, NotImplementedError):
      mesh_devices = np.reshape(jax.devices(), logical_mesh_shape)
  else:
    devices = np.empty(physical_mesh_shape, dtype=object)
    for device in jax.devices():
      devices[tuple(device.coords)] = device

    topology = 'x'.join(map(str, physical_mesh_shape))
    logical_mesh_shape = tuple(
        sizes[dim] for dim in ['ensemble', 'z', 'x', 'y']
    )
    rearrangement = _TPU_LAYOUT_REARRANGEMENTS[topology][logical_mesh_shape]

    abbreviated_sizes = {
        'e': sizes['ensemble'],
        'z': sizes['z'],
        'x': sizes['x'],
        'y': sizes['y'],
    }
    abbreviated_sizes = {k: v for k, v in abbreviated_sizes.items() if v != 1}
    mesh_devices = einops.rearrange(devices, rearrangement, **abbreviated_sizes)

  return jax.sharding.Mesh(mesh_devices, axis_names)


P = jax.sharding.PartitionSpec


def make_distributed_array_from_local_arrays(
    pytree: PyTree,
    mesh: jax.sharding.Mesh,
    spatial_partitions: jax.sharding.PartitionSpec,
    global_batch_size: int,
) -> PyTree:
  """Creates a pytree of global jax arrays for data/model parallelsm.

  This function exists for loading spatially partitioned data, which is assumed
  to be replicated across the ensemble dimension.

  Args:
    pytree: PyTree of NumPy arrays to convert into distributed JAX arrays. The
      leading "batch" dimension is divided between different local devices.
    mesh: SPDM sharding mesh.
    spatial_partitions: JAX partition spec (of length 3) to use for partitioning
      spatial dimensions (z, x, y).
    global_batch_size: number distinct examples in a single batch across all
      devices. Does not include the ensemble.

  Returns:
    Pytree with the same structure as the inputs, but with arrays replaced by
    distributed JAX arrays.
  """
  if len(spatial_partitions) != 3:
    raise ValueError(f'invalid {spatial_partitions=}')

  def get_shard_count(spec_part: None | str | tuple[str, ...]) -> int:
    # calculate the number of shards corresponding to an element in a
    # PartitionSpec
    if spec_part is None:
      return 1
    elif isinstance(spec_part, str):
      return mesh.shape[spec_part]
    else:
      return math.prod(mesh.shape[x] for x in spec_part)

  def shard_array(x: np.ndarray) -> jax.Array:
    if x.ndim <= 3:
      # handle sim_time [batch]
      global_shape = (global_batch_size,) + x.shape[1:]
      partition_spec = P('batch', *([None] * (x.ndim - 1)))
    elif x.ndim == 4:
      # This is currently needed to handle surface data that has shape:
      # [batch, time, x, y].
      _, x_shards, y_shards = map(get_shard_count, spatial_partitions)
      global_shape = (
          global_batch_size,
          x.shape[1],
          x.shape[2] * x_shards,
          x.shape[3] * y_shards,
      )
      partition_spec = P('batch', None, *spatial_partitions[1:])
    else:
      # everything else has dimensions [batch, time, z, x, y]
      assert x.ndim == 5, x.shape
      z_shards, x_shards, y_shards = map(get_shard_count, spatial_partitions)
      if x.shape[2] == 1:
        z_shards = 1
      global_shape = (
          global_batch_size,
          x.shape[1],
          x.shape[2] * z_shards,
          x.shape[3] * x_shards,
          x.shape[4] * y_shards,
      )
      partition_spec = P('batch', None, *spatial_partitions)

    sharding = jax.sharding.NamedSharding(mesh, partition_spec)
    single_device_arrays = put_to_devices(x, jax.local_devices(), axis=0)
    return jax.make_array_from_single_device_arrays(
        global_shape, sharding, single_device_arrays
    )

  try:
    return jax.tree_util.tree_map(shard_array, pytree)
  except Exception as e:
    shape_tree = jax.tree_util.tree_map(jnp.shape, pytree)
    raise RuntimeError(
        f'failed to shard arrays with shapes {shape_tree!r}'
    ) from e


def put_to_devices(
    host_array: np.ndarray, local_devices: abc.Sequence[Any], axis: int
) -> list[Any]:
  """Transfers a host array to local devices, split on the first dimension."""
  local_device_count = len(local_devices)
  try:
    per_device_arrays = np.split(host_array, local_device_count, axis=axis)
  except ValueError as array_split_error:
    raise ValueError(
        f'Unable to put to devices shape {host_array.shape} with '
        f'local device count {local_device_count}'
    ) from array_split_error
  device_buffers = [
      jax.device_put(arr, d) for arr, d in zip(per_device_arrays, local_devices)
  ]
  return device_buffers


def ensure_replicated(pytree: PyTree, *, mesh: jax.sharding.Mesh) -> PyTree:
  """Ensure that a pytree is replicated across all devices."""

  def replicate(x):
    x = jnp.asarray(x)
    spec = jax.sharding.PartitionSpec(*([None] * x.ndim))
    sharding = jax.sharding.NamedSharding(mesh, spec)
    return jax.lax.with_sharding_constraint(x, sharding)

  return jax.tree_util.tree_map(replicate, pytree)


T = TypeVar('T')


def jit_once(f: T, **jit_kwargs) -> T:
  """Like jax.jit, but raises an error instead of compiling multiple times."""
  compiled = None

  def g(*args, **kwargs):
    nonlocal compiled
    if compiled is None:
      logging.info(f'lowering {f}')
      lowered = jax.jit(f, **jit_kwargs).lower(*args, **kwargs)
      logging.info(f'compiling {f}')
      compiled = lowered.compile()
      logging.info(f'finishing compiling {f}')
    return compiled(*args, **kwargs)

  return g



# Copyright 2024 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests that pytree_transforms work as expected and  pytrees with expected shapes."""

import functools
from typing import Any, Sequence

from absl.testing import absltest
from absl.testing import parameterized
import chex
from flax import nnx
import jax
import jax.numpy as jnp
from neuralgcm.experimental import coordax as cx
from neuralgcm.experimental import coordinates
from neuralgcm.experimental import data_specs
from neuralgcm.experimental import dynamic_io
from neuralgcm.experimental import orographies
from neuralgcm.experimental import pytree_mappings
from neuralgcm.experimental import pytree_transforms
from neuralgcm.experimental import pytree_utils
from neuralgcm.experimental import random_processes
from neuralgcm.experimental import standard_layers
from neuralgcm.experimental import towers
from neuralgcm.experimental import typing
from neuralgcm.experimental import units
import numpy as np
import xarray


class StandardPytreeTransformsTest(parameterized.TestCase):
  """Tests for standard_layers.MaskTransform and Composed.ConvLonLat."""

  @parameterized.parameters(
      dict(
          fields_to_mask=['u', 'v', 'T'],
          mask_shape=(5, 5),
          fill_value_true=0.0,
          fill_value_false=1.0,
          fill_threshold=jnp.nan,
      ),
      dict(
          fields_to_mask=['T'],
          mask_shape=(5, 5),
          fill_value_true=0.0,
          fill_value_false=1.0,
          fill_threshold=0.1,
      ),
      dict(
          fields_to_mask=['T'],
          mask_shape=(5, 5),
          fill_value_true=0.0,
          fill_value_false=1.0,
          fill_threshold=0.3,
      ),
  )
  def test_mask_transform(
      self,
      fields_to_mask: Sequence[str],
      mask_shape: tuple[int, int],
      fill_value_true: float,
      fill_value_false: float,
      fill_threshold: float,
  ):
    if fill_threshold is np.nan:
      nan_threshold = True
      threshold = 0.7
    else:
      nan_threshold = False
      threshold = fill_threshold
    data = {
        'u': ((np.arange(np.prod(mask_shape)) % 10) / 10).reshape(mask_shape),
        'v': ((np.arange(np.prod(mask_shape)) % 10) / 10).reshape(mask_shape),
        'T': ((np.arange(np.prod(mask_shape)) % 10) / 10).reshape(mask_shape),
    }
    expected = {
        k: (
            np.where(v > threshold, fill_value_true, fill_value_false)
            if k in fields_to_mask
            else v
        )
        for k, v in data.items()
    }
    if nan_threshold:
      inputs = {k: np.where(v > threshold, np.nan, v) for k, v in data.items()}
    else:
      inputs = data
    if len(mask_shape) == 2:
      xarray_test = xarray.Dataset(
          data_vars=dict(
              u=(['longitude', 'latitude'], inputs['u']),
              v=(['longitude', 'latitude'], inputs['v']),
              T=(['longitude', 'latitude'], inputs['T']),
          ),
          coords=dict(
              longitude=('longitude', np.arange(mask_shape[-2])),
              latitude=('latitude', np.arange(mask_shape[-1])),
          ),
      )
    elif len(mask_shape) == 3:
      xarray_test = xarray.Dataset(
          data_vars=dict(
              u=(['longitude', 'latitude', 'level'], inputs['u']),
              v=(['longitude', 'latitude', 'level'], inputs['v']),
              T=(['longitude', 'latitude', 'level'], inputs['T']),
          ),
          coords=dict(
              longitude=('longitude', np.arange(mask_shape[-3])),
              latitude=('latitude', np.arange(mask_shape[-2])),
              level=('level', np.arange(mask_shape[-1])),
          ),
      )
    else:
      raise ValueError(f'Expected mask_shape of length 2 or 3. {mask_shape=}')
    mask_transform = pytree_transforms.MaskTransform(
        fields_to_mask=fields_to_mask,
        mask_shape=mask_shape,
        fill_value_true=fill_value_true,
        fill_value_false=fill_value_false,
        fill_threshold=fill_threshold,
    )

    with self.subTest('gen_mask_from_xarray'):
      mask_transform.update_from_xarray(xarray_test, fields_to_mask[0])
      np.testing.assert_allclose(
          mask_transform.mask.value, expected[fields_to_mask[0]]
      )

    test_data = {
        'u': np.ones(mask_shape),
        'v': np.ones(mask_shape),
        'T': np.ones(mask_shape),
    }
    with self.subTest('check_mask_applies_to_correct_variables'):
      outputs = mask_transform(test_data)
      for i in set(['u', 'v', 'T']) - set(fields_to_mask):
        np.testing.assert_allclose(outputs[i], np.ones(mask_shape))
      for i in fields_to_mask:
        np.testing.assert_allclose(outputs[i], expected[i])

  @parameterized.parameters(
      dict(
          input_dict={'a': 3, 'b': 4, 'c': 5},
          keys_to_nest=('b',),
          nested_key_name='nested_b',
          expected={'a': 3, 'nested_b': {'b': 4}, 'c': 5},
      ),
      dict(
          input_dict={'a': 3, 'b': 4, 'c': 5, '6': 6},
          keys_to_nest=('b', 'c', '6'),
          nested_key_name='tracers',
          expected={'a': 3, 'tracers': {'b': 4, 'c': 5, '6': 6}},
      ),
  )
  def test_nest_dict_transform(
      self,
      input_dict: Sequence[str],
      keys_to_nest: tuple[str, ...],
      nested_key_name: str,
      expected: dict[str, Any],
  ):
    nest_dict_transform = pytree_transforms.NestDict(
        keys_to_nest=keys_to_nest, nested_key_name=nested_key_name
    )
    actual = nest_dict_transform(input_dict)
    chex.assert_trees_all_equal(actual, expected)

  @parameterized.parameters(
      dict(n_clip=1),
      dict(n_clip=2),
      dict(n_clip=5),
  )
  def test_clip_wavenumbers(self, n_clip: int = 1):
    """Tests that ClipWavenumbers works as expected."""
    grid = coordinates.SphericalHarmonicGrid.T21()
    inputs = {
        'u': np.ones(grid.shape),
        'v': np.ones(grid.shape),
    }
    ls = grid.fields['total_wavenumber'].data
    clip_mask = (np.arange(ls.size) <= (ls.max() - n_clip)).astype(int)
    expected = jax.tree.map(lambda x: x * clip_mask, inputs)
    clip_transform = pytree_transforms.ClipWavenumbers(
        grid=grid, wavenumbers_to_clip=n_clip
    )
    actual = clip_transform(inputs)
    chex.assert_trees_all_equal(actual, expected)

  def test_batch_shift_and_normalize(self):
    """Tests that BatchShiftAndNormalize works as expected."""
    feature_size = 2
    batch_size = 200
    shape = (batch_size, feature_size)

    keys = ('a', 'b', 'c', 'd')
    input_means = (2.0, 0.0, -12.5, 100.0)
    input_stds = (0.2, 1.5, 3.14, 100.0)

    def get_inputs(rng):
      rngs = jax.random.split(rng, len(keys))
      return {
          k: jax.random.normal(rng, shape=shape) * std + mean
          for k, rng, mean, std in zip(keys, rngs, input_means, input_stds)
      }

    xs = get_inputs(jax.random.PRNGKey(1))

    def _check_mean_and_std(xs, expected_means, expected_stds):
      mean_over_batch = functools.partial(np.mean, axis=0)
      std_over_batch = functools.partial(np.std, axis=0)
      xs_mean = jax.tree.map(mean_over_batch, xs)
      xs_std = jax.tree.map(std_over_batch, xs)
      for i, k in enumerate(keys):
        mean_atol = 6 * (expected_stds[i] / np.sqrt(batch_size))
        std_atol = 6 * (np.sqrt(2 / (batch_size - 1)) * expected_stds[i]**2)
        expected_mean = np.array([expected_means[i]] * feature_size)
        expected_std = np.array([expected_stds[i]] * feature_size)
        np.testing.assert_allclose(xs_mean[k], expected_mean, atol=mean_atol)
        np.testing.assert_allclose(xs_std[k], expected_std, atol=std_atol)

    with self.subTest('input_mean_and_std'):
      _check_mean_and_std(xs, input_means, input_stds)

    batch_shift_and_normalize = pytree_transforms.BatchShiftAndNormalize(
        {k: feature_size for k in keys},
        feature_axis=-1,
        momentum=0.1,
        use_running_average=True,
    )
    with self.subTest('identity_at_init'):
      ys = batch_shift_and_normalize(xs)
      _check_mean_and_std(ys, input_means, input_stds)

    zero_means = tuple(0.0 for _ in input_means)
    unit_stds = tuple(1.0 for _ in input_stds)
    with self.subTest('zero_mean_unit_variance_when_dynamic'):
      batch_shift_and_normalize.use_running_average = False
      ys = batch_shift_and_normalize(xs)
      _check_mean_and_std(ys, zero_means, unit_stds)

    with self.subTest('converges_to_zero_mean_unit_variance'):
      for _ in range(20):  # EMA converges with remaining init bias ~0.1**20.
        _ = batch_shift_and_normalize(xs)
      batch_shift_and_normalize.use_running_average = True
      ys = batch_shift_and_normalize(xs)
      _check_mean_and_std(ys, zero_means, unit_stds)


class InputsFeaturesTest(parameterized.TestCase):
  """Tests input features modules."""

  def _test_feature_module(
      self,
      feature_module: pytree_transforms.Transform,
      inputs: typing.Pytree,
  ):
    with self.subTest('call method'):
      features = feature_module(inputs)

    with self.subTest('output_shape_matches_actual'):
      input_shapes = pytree_utils.shape_structure(inputs)
      actual = feature_module.output_shapes(input_shapes)
      expected = pytree_utils.shape_structure(features)
      chex.assert_trees_all_equal(actual, expected)

  @parameterized.named_parameters(
      dict(
          testcase_name='dinosaur_coords',
          coords=coordinates.DinosaurCoordinates(
              horizontal=coordinates.LonLatGrid.T21(),
              vertical=coordinates.SigmaLevels.equidistant(4),
          ),
      ),
  )
  def test_radiation_features(self, coords):
    radiation_features = pytree_transforms.RadiationFeatures(
        coords=coords,
        sim_units=units.DEFAULT_UNITS,
    )
    self._test_feature_module(
        radiation_features,
        {'sim_time': np.array(314.159)},
    )

  def test_latitude_features(self):
    grid = coordinates.LonLatGrid.T21()
    latitude_features = pytree_transforms.LatitudeFeatures(grid=grid)
    self._test_feature_module(latitude_features, None)

  def test_orography_features(self):
    grid = coordinates.SphericalHarmonicGrid.T21()
    orography = orographies.ModalOrography(grid=grid, rngs=None)
    orography_features = pytree_transforms.OrographyFeatures(
        orography_module=orography,
    )
    self._test_feature_module(orography_features, None)

  def test_orography_with_grads_features(self):
    grid = coordinates.SphericalHarmonicGrid.T21()
    orography = orographies.ModalOrography(grid=grid, rngs=None)
    orography_features = pytree_transforms.OrographyWithGradsFeatures(
        orography_module=orography,
        compute_gradients_transform=pytree_transforms.ToModalWithFilteredGradients(
            grid, filter_attenuations=[2.0]
        ),
    )
    self._test_feature_module(orography_features, None)

  def test_dynamic_input_features(self):
    grid = coordinates.LonLatGrid.T21()
    dynamic_input = dynamic_io.DynamicInputSlice(
        keys_to_coords={'a': grid, 'b': grid, 'c': grid}
    )
    expand_dims = lambda x: np.expand_dims(x, axis=(1, 2))
    data = {
        'a': expand_dims(np.arange(2)) * np.ones(grid.shape),
        'b': expand_dims(np.arange(2)) * np.zeros(grid.shape),
        'c': expand_dims(np.arange(2)) * np.ones(grid.shape),
    }
    timedelta = coordinates.TimeDelta.as_index(2)
    grid_trajectory = cx.compose_coordinates(timedelta, grid)
    sim_time = np.arange(timedelta.shape[0])
    in_data = {
        k: data_specs.TimedField(cx.wrap(v, grid_trajectory), sim_time)
        for k, v in data.items()
    }
    dynamic_input.update_dynamic_inputs(in_data)
    with self.subTest('two_keys'):
      dynamic_input_features = pytree_transforms.DynamicInputFeatures(
          ('a', 'b'), dynamic_input
      )
      self._test_feature_module(dynamic_input_features, {'sim_time': 0.1})

  def test_spatial_surface_features(self):
    feature_sizes = {
        'learned_surface_features': 8,
    }
    grid = coordinates.LonLatGrid.T21()
    static_surface_features = pytree_transforms.SpatialSurfaceFeatures(
        feature_sizes, grid=grid, rngs=nnx.Rngs(1)
    )
    self._test_feature_module(static_surface_features, None)

  def test_velocity_and_prognostics_with_modal_gradients(self):
    coords = coordinates.DinosaurCoordinates(
        horizontal=coordinates.LonLatGrid.T21(),
        vertical=coordinates.SigmaLevels.equidistant(3),
    )
    with_gradients_transform = pytree_transforms.ToModalWithFilteredGradients(
        coords.horizontal, filter_attenuations=[2.0]
    )
    features_grads = pytree_transforms.VelocityAndPrognosticsWithModalGradients(
        coords,
        volume_field_names=(
            'u',
            'v',
        ),
        surface_field_names=('lsp',),
        compute_gradients_transform=with_gradients_transform,
    )
    inputs = {
        'u': np.ones(coords.dinosaur_coords.modal_shape),
        'v': np.ones(coords.dinosaur_coords.modal_shape),
        'lsp': np.ones(coords.dinosaur_coords.modal_shape[1:])[np.newaxis, ...],
        'tracers': {},
        'sim_time': 0.0,
    }
    self._test_feature_module(features_grads, inputs)

  def test_surface_embedding_features(self):
    n_levels = 12
    coords = coordinates.DinosaurCoordinates(
        horizontal=coordinates.LonLatGrid.T21(),
        vertical=coordinates.LayerLevels(n_levels),
    )
    mlp_factory = functools.partial(
        standard_layers.MlpUniform, hidden_size=6, n_hidden_layers=2
    )
    tower_factory = functools.partial(
        towers.ColumnTower, column_net_factory=mlp_factory
    )
    mapping_factory = functools.partial(
        pytree_mappings.ChannelMapping,
        tower_factory=tower_factory,
    )
    feature_module = pytree_transforms.LatitudeFeatures(
        grid=coords.horizontal,
    )
    embedding_factory = functools.partial(
        pytree_mappings.Embedding,
        feature_module=feature_module,
        mapping_factory=mapping_factory,
        rngs=nnx.Rngs(0),
    )
    surface_embedding_features = pytree_transforms.SurfaceEmbeddingFeatures(
        coords=coords,
        embedding_sizes={'abc': 3, 'foo': 5},
        embedding_factory=embedding_factory,
    )
    self._test_feature_module(surface_embedding_features, None)

  def test_volume_embedding_features(self):
    n_levels = 12
    coords = coordinates.DinosaurCoordinates(
        horizontal=coordinates.LonLatGrid.T21(),
        vertical=coordinates.LayerLevels(n_levels),
    )
    mlp_factory = functools.partial(
        standard_layers.MlpUniform, hidden_size=6, n_hidden_layers=2
    )
    tower_factory = functools.partial(
        towers.ColumnTower, column_net_factory=mlp_factory
    )
    mapping_factory = functools.partial(
        pytree_mappings.ChannelMapping,
        tower_factory=tower_factory,
    )
    feature_module = pytree_transforms.LatitudeFeatures(
        grid=coords.horizontal,
    )
    embedding_factory = functools.partial(
        pytree_mappings.Embedding,
        feature_module=feature_module,
        mapping_factory=mapping_factory,
        rngs=nnx.Rngs(0),
    )
    surface_embedding_features = pytree_transforms.VolumeEmbeddingFeatures(
        coords=coords,
        embedding_names=('abc', 'fuz', 'bar'),
        embedding_factory=embedding_factory,
    )
    self._test_feature_module(surface_embedding_features, None)

  @parameterized.named_parameters(
      dict(
          testcase_name='T21_grid',
          grid=coordinates.LonLatGrid.T21(),
      ),
  )
  def test_randomness_features(self, grid):
    with self.subTest('gaussian_random_field'):
      random_process = random_processes.GaussianRandomField(
          grid=grid,  # instantiating here because uses jax at __init__.
          dt=1.0,
          sim_units=units.DEFAULT_UNITS,
          correlation_time=1.0,
          correlation_length=1.0,
          variance=1.0,
          rngs=nnx.Rngs(0),
      )
      random_process.unconditional_sample(jax.random.PRNGKey(0))
      randomness_features = pytree_transforms.RandomnessFeatures(
          random_process=random_process,
          grid=grid,
      )
      self._test_feature_module(randomness_features, None)

    with self.subTest('batched_gaussian_random_fields'):
      random_process = random_processes.BatchGaussianRandomField(
          grid=grid,  # instantiating here because uses jax at __init__.
          dt=1.0,
          sim_units=units.DEFAULT_UNITS,
          correlation_times=[1.0, 2.0],
          correlation_lengths=[0.6, 0.9],
          variances=[1.0, 1.0],
          rngs=nnx.Rngs(0),
      )
      random_process.unconditional_sample(jax.random.PRNGKey(0))
      randomness_features = pytree_transforms.RandomnessFeatures(
          random_process=random_process,
          grid=grid,
      )
      self._test_feature_module(randomness_features, None)

  def test_prognostic_features(self):
    coords = coordinates.DinosaurCoordinates(
        horizontal=coordinates.LonLatGrid.T21(),
        vertical=coordinates.LayerLevels(n_layers=3),
    )
    prognostic_features = pytree_transforms.PrognosticFeatures(
        coords=coords, surface_field_names=('a', 'b'), volume_field_names=('c',)
    )
    inputs = {
        'a': np.ones((1,) + coords.horizontal.shape),
        'b': np.ones((1,) + coords.horizontal.shape),
        'c': np.ones(coords.shape),
        'sim_time': 0.0,
    }
    self._test_feature_module(prognostic_features, inputs)

  def test_pressure_features(self):
    coords = coordinates.DinosaurCoordinates(
        horizontal=coordinates.SphericalHarmonicGrid.T21(),
        vertical=coordinates.SigmaLevels.equidistant(8),
    )
    pressure_features = pytree_transforms.PressureFeatures(coords=coords)
    inputs = {
        'log_surface_pressure': np.ones((1,) + coords.horizontal.shape),
    }
    self._test_feature_module(pressure_features, inputs)

if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configurable optimizers from JAX."""
import collections
import re
from typing import Sequence

import gin
import optax


gin.external_configurable(optax.adabelief, module='optax')
gin.external_configurable(optax.adam, module='optax')
gin.external_configurable(optax.adamw, module='optax')

gin.external_configurable(optax.constant_schedule, module='optax')
gin.external_configurable(optax.join_schedules, module='optax')
gin.external_configurable(optax.piecewise_constant_schedule, module='optax')
gin.external_configurable(optax.exponential_decay, module='optax')
gin.external_configurable(
    optax.warmup_exponential_decay_schedule, module='optax'
)


class OptimizerError(Exception):
  """Raised if a custom Whirl optimizer encounters an error."""


@gin.configurable
def optimizer(value):
  return value


OptState = collections.namedtuple('OptState', ['state', 'params'])


@gin.register
def piecewise_constant_schedule_specified_by_rates(
    rates: Sequence[float],
    boundaries: Sequence[int],
) -> optax.Schedule:
  """Schedule that is piecewise constant and specified by rates (not scales).

  This is similar to optax.piecewise_constant_schedule, which requires users
  to specify "scales" (ratio of old LR to new LR).

  Args:
    rates: Length K sequence of learning rates. `rates[i]` is used for steps
      `0 <= step < boundaries[1]`, for i=0, and
      `boundaries[i-1] <= step < boundaries[i]`, for 0 < i < len(boundaries)
      `boundaries[i-1] <= step < ∞`, for i = len(boundaries)
    boundaries: Length K-1 sequence of boundaries.

  Returns:
    Schedule to pass to optax optimizers.
  """
  return optax.join_schedules(
      schedules=[optax.constant_schedule(r) for r in rates],
      boundaries=boundaries,
  )


@gin.register
def delayed_constant_schedule(
    turn_on_step: int,
    rate: float,
) -> optax.Schedule:
  """Schedule that is zero until `turn_on_step` then `rate` thereafter."""
  return piecewise_constant_schedule_specified_by_rates(
      rates=[0., rate],
      boundaries=[turn_on_step],
  )


@gin.register
def top_level_multi_adam(
    top_level_keys: Sequence[str] = (),
    learning_rates: Sequence[optax.ScalarOrSchedule] = (),
    default_learning_rate: optax.ScalarOrSchedule = 1e-4,
    b1: float = 0.9,
    b2: float = 0.95,
    eps: float = 1e-6,
    raise_if_keys_not_found: bool = True,
) -> optax.GradientTransformation:
  """Uses an Adam optimizer with different learning rates for different params.

  Args:
    top_level_keys: Keys to use non-default learning rates for. A key starting
      with 'REGEX_', such as 'REGEX_cats' will use re.search to find keys, e.g.
      re.search('cats', key).
    learning_rates: Learning rates to use leafs under the `top_level_keys`.
    default_learning_rate: Learning rate to use for keys not in `learning_rates`
    b1: Exponential decay to track the first moment of past gradients.
    b2: Exponential decay to track the second moment of past gradients.
    eps:  A small constant applied to denominator outside of the square root to
      avoid dividing by zero when rescaling.
    raise_if_keys_not_found: Whether to raise if some `top_level_keys` are not
      found in params.

  Returns:
    optax optimizer with learning rate based on top level key in params dict.
  """
  if len(top_level_keys) != len(learning_rates):
    raise ValueError(
        f'{top_level_keys=} had different length than {learning_rates=}'
    )
  if '' in top_level_keys:
    raise ValueError('An empty string "" was found in `top_level_keys`.')

  default_label = 'DEFAULT_LABEL'
  if default_label in top_level_keys:
    raise ValueError(f'{default_label=} should not be in `top_level_keys`')

  def find_matching_top_level_key(param_name: str) -> str:
    """Searches for param_name in top_level_keys, returns the matching key."""
    prefix = 'REGEX_'
    matches = []
    for k in top_level_keys:
      if k.startswith(prefix) and re.search(k.lstrip(prefix), param_name):
        matches.append(k)
      elif k == param_name:
        matches.append(k)
    if not matches:
      return default_label
    elif len(matches) == 1:
      return matches[0]
    else:
      raise ValueError(
          f'{param_name=} had more than 1 ({len(matches)}) match '
          f'({matches}). Only one `top_level_keys` should match, or else we '
          'cannot choose a unique learning rate for these parameters.'
      )

  def get_prefix_labels(params):
    """Makes prefix labels to help optax match params with learning rates."""
    # E.g. if top_level_keys = ['module_A', 'REGEX_special'],
    # and params.keys() = ['module_A', 'special_A', 'special_B', 'module_C'],
    # labels = {
    #   'module_A': 'module_A',
    #   'special_A': 'REGEX_special', 'special_B': 'REGEX_special',
    #   'module_C': 'DEFAULT_LABEL', 'module_D': 'DEFAULT_LABEL',...
    # }
    # E.g. labels tells optax to use the learning rate 'REGEX_special' for
    # parameters under the prefix 'module_C'.
    labels = {
        param_name: find_matching_top_level_key(param_name)
        for param_name in params
    }
    top_level_keys_that_matched = [
        k for k in labels.values() if k != default_label
    ]
    missing_keys = set(top_level_keys).difference(top_level_keys_that_matched)
    if raise_if_keys_not_found and missing_keys:
      raise OptimizerError(
          f'{missing_keys=} not found in params: {sorted(params)}'
      )
    return labels

  def make_adam(lr):
    return optax.adam(lr, b1=b1, b2=b2, eps=eps)

  return optax.multi_transform(
      transforms={
          k: make_adam(lr) for k, lr in zip(top_level_keys, learning_rates)
      }
      | {default_label: make_adam(default_learning_rate)},
      param_labels=get_prefix_labels,
  )



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import importlib.resources
import pickle

from dinosaur import coordinate_systems
from dinosaur import horizontal_interpolation
from dinosaur import spherical_harmonic
import neuralgcm
import numpy as np
import xarray


def _horizontal_regrid(
    regridder: horizontal_interpolation.Regridder, dataset: xarray.Dataset
) -> xarray.Dataset:
  """Horizontally regrid an xarray Dataset."""
  # TODO(shoyer): consider moving to public API
  regridded = xarray.apply_ufunc(
      regridder,
      dataset,
      input_core_dims=[['longitude', 'latitude']],
      output_core_dims=[['longitude', 'latitude']],
      exclude_dims={'longitude', 'latitude'},
      vectorize=True,  # loops over level, for lower memory usage
  )
  regridded.coords['longitude'] = np.rad2deg(regridder.target_grid.longitudes)
  regridded.coords['latitude'] = np.rad2deg(regridder.target_grid.latitudes)
  return regridded


def load_checkpoint_tl63_stochastic():
  """Load a checkpoint for a toy TL63 stochastic model."""
  package = importlib.resources.files(neuralgcm)
  file = package.joinpath('data/tl63_stochastic_mini.pkl')
  return pickle.loads(file.read_bytes())


def load_data(coords: coordinate_systems.CoordinateSystem) -> xarray.Dataset:
  """Load demo data for the given coordinate system."""
  if coords.vertical.layers != 37:
    raise ValueError('can only load demo data for 37 pressure levels')
  package = importlib.resources.files(neuralgcm)
  with package.joinpath('data/era5_tl31_19590102T00.nc').open('rb') as f:
    ds = xarray.load_dataset(f).expand_dims('time')
  regridder = horizontal_interpolation.ConservativeRegridder(
      spherical_harmonic.Grid.TL31(), coords.horizontal
  )
  return _horizontal_regrid(regridder, ds)



# Copyright 2024 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Modules that hold orographic data."""

from flax import nnx
from neuralgcm.experimental import coordax as cx
from neuralgcm.experimental import coordinates
from neuralgcm.experimental import interpolators
from neuralgcm.experimental import spatial_filters
from neuralgcm.experimental import typing
from neuralgcm.experimental import units
from neuralgcm.experimental import xarray_utils
import xarray


class OrographyVariable(nnx.Variable):
  """Variable class for orography data."""


class ModalOrography(nnx.Module):
  """Orogrphay module that provoides elevation in modal representation."""

  def __init__(
      self,
      *,
      grid: coordinates.SphericalHarmonicGrid,
      initializer: nnx.initializers.Initializer = nnx.initializers.zeros_init(),
      rngs: nnx.Rngs,
  ):
    self.grid = grid
    self.orography = OrographyVariable(
        initializer(rngs, grid.ylm_grid.modal_shape))

  @property
  def nodal_orography(self) -> typing.Array:
    return self.grid.ylm_grid.to_nodal(self.orography.value)

  @property
  def modal_orography(self) -> typing.Array:
    """Returns orography converted to modal representation with filtering."""
    return self.orography.value

  def update_orography_from_data(
      self,
      dataset: xarray.Dataset,
      sim_units: units.SimUnits,
      spatial_filter=None,
  ):
    """Updates ``self.orography`` with filtered orography from dataset."""
    # TODO(dkochkov) use units attr on dataset with default to `meter` here.
    if spatial_filter is None:
      spatial_filter = lambda x: x
    nodal_orography = xarray_utils.nodal_orography_from_ds(dataset)
    nodal_orography = xarray_utils.xarray_nondimensionalize(
        nodal_orography, sim_units
    ).values
    data_coords = xarray_utils.coordinates_from_dataset(dataset)
    assert isinstance(data_coords, coordinates.DinosaurCoordinates)
    if not isinstance(spatial_filter, spatial_filters.ModalSpatialFilter):
      nodal_orography = spatial_filter(nodal_orography)
    data_grid = data_coords.horizontal
    if isinstance(data_grid, coordinates.SphericalHarmonicGrid):
      data_modal_grid = data_grid
    elif isinstance(data_grid, coordinates.LonLatGrid):
      data_modal_grid = data_grid.to_spherical_harmonic_grid()
    else:
      raise ValueError(f'Unsupported data grid {data_grid=}')
    modal_orography = data_modal_grid.ylm_grid.to_modal(nodal_orography)
    interpolator = interpolators.SpectralRegridder(self.grid)
    modal_orography = interpolator(cx.wrap(modal_orography, data_modal_grid))
    modal_orography = modal_orography.unwrap(self.grid)
    if isinstance(spatial_filter, spatial_filters.ModalSpatialFilter):
      modal_orography = spatial_filter.filter_modal(modal_orography)
    self.orography.value = modal_orography


class Orography(nnx.Module):
  """Orography module that provides elevation in real space."""

  def __init__(
      self,
      *,
      grid: coordinates.LonLatGrid,
      initializer: nnx.initializers.Initializer = nnx.initializers.zeros_init(),
      rngs: nnx.Rngs,
  ):
    self.grid = grid
    self.orography = OrographyVariable(
        initializer(rngs, grid.ylm_grid.nodal_shape)
    )

  @property
  def nodal_orography(self) -> typing.Array:
    return self.orography.value

  def update_orography_from_data(
      self,
      dataset: xarray.Dataset,
      sim_units: units.SimUnits,
      spatial_filter=None,
  ):
    """Updates ``self.orography`` with filtered orography from dataset."""
    # TODO(dkochkov) use units attr on dataset with default to `meter` here.
    if spatial_filter is None:
      spatial_filter = lambda x: x
    nodal_orography = xarray_utils.nodal_orography_from_ds(dataset)
    nodal_orography = xarray_utils.xarray_nondimensionalize(
        nodal_orography, sim_units
    ).values
    data_coords = xarray_utils.coordinates_from_dataset(dataset)
    assert isinstance(data_coords, coordinates.DinosaurCoordinates)
    nodal_orography = spatial_filter(nodal_orography)
    data_grid = data_coords.horizontal
    if data_grid != self.grid:
      raise ValueError(f'{data_grid=} does not match {self.grid=}.')
    self.orography.value = nodal_orography



# Copyright 2024 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Modules that implement spatial filters."""

import abc
import dataclasses
from typing import Sequence

from dinosaur import filtering
from neuralgcm.experimental import coordinates
from neuralgcm.experimental import typing
from neuralgcm.experimental import units
import numpy as np


class SpatialFilter(abc.ABC):
  """Base class for spatial filters."""

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    """Returns filtered ``inputs``."""


class ModalSpatialFilter(SpatialFilter):
  """Base class for filters."""

  @abc.abstractmethod
  def filter_modal(self, inputs: typing.Pytree) -> typing.Pytree:
    """Returns filtered modal ``inputs``."""


class ExponentialModalFilter(ModalSpatialFilter):
  """Modal filter that removes high frequency components."""

  def __init__(
      self,
      grid: coordinates.SphericalHarmonicGrid | coordinates.LonLatGrid,
      attenuation: float = 16,
      order: int = 18,
      cutoff: float = 0,
  ):
    """See ``dinosaur.filtering.exponential_filter`` for details."""
    self.grid = grid
    self.filter = filtering.exponential_filter(
        grid.ylm_grid, attenuation, order, cutoff)

  def filter_modal(self, inputs: typing.Pytree) -> typing.Pytree:
    return self.filter(inputs)

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    dinosaur_grid = self.grid.ylm_grid
    return dinosaur_grid.to_nodal(self.filter(dinosaur_grid.to_modal(inputs)))

  @classmethod
  def from_timescale(
      cls,
      grid: coordinates.SphericalHarmonicGrid | coordinates.LonLatGrid,
      dt: float | typing.Quantity | typing.Numeric,
      timescale: float | typing.Quantity | typing.Numeric,
      order: int = 18,
      cutoff: float = 0,
      *,
      sim_units: units.SimUnits,
  ):
    """Returns a filter with the given timescale."""
    if isinstance(dt, np.timedelta64):
      dt = units.nondimensionalize_timedelta64(dt, sim_units)
    else:
      dt = units.maybe_nondimensionalize(dt, sim_units)
    if isinstance(timescale, np.timedelta64):
      timescale = units.nondimensionalize_timedelta64(timescale, sim_units)
    else:
      timescale = units.maybe_nondimensionalize(timescale, sim_units)
    return cls(grid, attenuation=(dt / timescale), order=order, cutoff=cutoff)


@dataclasses.dataclass
class SequentialModalFilter(ModalSpatialFilter):
  """Modal filter that applies multiple filters sequentially."""

  filters: Sequence[ModalSpatialFilter]
  grid: coordinates.SphericalHarmonicGrid | coordinates.LonLatGrid

  def filter_modal(self, inputs: typing.Pytree) -> typing.Pytree:
    for modal_filter in self.filters:
      inputs = modal_filter.filter_modal(inputs)
    return inputs

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    dinosaur_grid = self.grid.ylm_grid
    modal_inputs = dinosaur_grid.to_modal(inputs)
    modal_outputs = self.filter_modal(modal_inputs)
    return dinosaur_grid.to_nodal(modal_outputs)



# Copyright 2024 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Modules that define learnable mappings between input/output pytrees."""

from typing import Callable, Protocol, Sequence

from flax import nnx
import jax

from neuralgcm.experimental import coordinates
from neuralgcm.experimental import pytree_transforms
from neuralgcm.experimental import pytree_utils
from neuralgcm.experimental import towers
from neuralgcm.experimental import typing


class PytreeMapping(Protocol):
  """Protocol for pytree_mapping modules."""

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    ...

  @property
  def output_shapes(self) -> typing.Pytree:
    ...


PytreeMappingFactory = Callable[..., PytreeMapping]


MINIMAL_STATE_STRUCT = {'sim_time': 0.0}


class ChannelMapping(nnx.Module):
  """Maps the pytree of nodal features to a pytree of specified structure.

  This module packs the pytree into a single array of shape (n, lon, lat),
  passes it to a NN tower, and unpacks the result into a pytree with the
  structure of output_shapes, typically (m, lon, lat).
  """

  def __init__(
      self,
      input_shapes: typing.Pytree,
      output_shapes: typing.Pytree,
      *,
      tower_factory: towers.UnaryTowerFactory,
      rngs: nnx.Rngs,
  ):
    # TODO(dkochkov) consider passing `coords` to infer how input features are
    # structured and which axis in output_shapes is a channel axis.
    f_axis = -3  # default column axis.
    input_size = sum([x.shape[f_axis] for x in jax.tree.leaves(input_shapes)])
    output_size = sum([x.shape[f_axis] for x in jax.tree.leaves(output_shapes)])
    # tower preserves the last two spatial dimensions.
    self.tower = tower_factory(input_size, output_size, rngs=rngs)
    self._output_shapes = output_shapes
    self.feature_axis = f_axis

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    # TODO(dkochkov) Consider adding check that inputs is not just an array.
    array = pytree_utils.pack_pytree(inputs, self.feature_axis)
    if array.ndim != 3:
      raise ValueError(f'Expected input array with ndim=3, got {array.shape=}')
    outputs = self.tower(array)
    if outputs.ndim != 3:
      raise ValueError(f'Expected outputs with ndim=3, got {outputs.shape=}')
    return pytree_utils.unpack_to_pytree(
        outputs, self._output_shapes, self.feature_axis
    )

  @property
  def output_shapes(self):
    return self._output_shapes


class VariableMapping(nnx.Module):
  """Maps the pytree of nodal volume features to a pytree of given structure.

  This module stacks the input pytree into an array of shape
  (channel, level, lon, lat), passes it to a NN tower.  The output from the NN
  is expected to have shape (n, level, lon, lat), and gets unpacked to a pytree
  with the structure of output_shapes, e.g.
    output_shapes = {
        'var_1': jnp.asarray((level, lon, lat)),
        'var_2': jnp.asarray((level, lon, lat)),
        ...,
        'var_n': jnp.asarray((level, lon, lat)),
    }

  The leaves of the input pytree must have the same shape, e.g. (1, lon, lat) or
  (level, lon, lat). To mix shapes, broadcast before passing to the mapping.
  """

  def __init__(
      self,
      input_shapes: typing.Pytree,
      output_shapes: typing.Pytree,
      *,
      tower_factory: towers.UnaryTowerFactory,
      rngs: nnx.Rngs,
  ):
    feature_axis = 0
    input_size = len(jax.tree.leaves(input_shapes))
    output_size = len(jax.tree.leaves(output_shapes))

    # tower preserves the last two spatial dimensions.
    self.tower = tower_factory(input_size, output_size, rngs=rngs)
    self._output_shapes = output_shapes
    self.feature_axis = feature_axis

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    array = pytree_utils.stack_pytree(inputs, axis=self.feature_axis)
    if array.ndim != 4:
      raise ValueError(f'Expected input array with ndim=4, got {array.shape=}')
    outputs = self.tower(array)
    if outputs.ndim != 4:
      raise ValueError(f'Expected outputs with ndim=4, got {outputs.shape=}')
    return pytree_utils.unstack_to_pytree(
        outputs, self._output_shapes, axis=self.feature_axis
    )

  @property
  def output_shapes(self):
    return self._output_shapes


class MappingWithNormalizedInputs(nnx.Module):
  """Applies pytree mapping with input normalization."""

  def __init__(
      self,
      input_shapes: typing.Pytree,
      output_shapes: typing.Pytree,
      *,
      mapping_factory: PytreeMappingFactory,
      normalization_factory: pytree_transforms.TransformFactory,
      rngs: nnx.Rngs,
  ):
    self._output_shapes = output_shapes
    self.normalization = normalization_factory(input_shapes, rngs=rngs)
    self.mapping = mapping_factory(
        input_shapes=self.normalization.output_shapes(input_shapes),
        output_shapes=output_shapes,
        rngs=rngs,
    )

  @property
  def output_shapes(self):
    return self._output_shapes

  def normalize(self, inputs: typing.Pytree) -> typing.Pytree:
    return self.normalization(inputs)

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    return self.mapping(self.normalize(inputs))


# TODO(dkochkov) Consider parameterizing this with mapping factories.
class ParallelMapping(nnx.Module):
  """Maps a pytree to a pytree by additively compbining multiple mappings."""

  def __init__(
      self,
      mappings: Sequence[PytreeMapping],
  ):
    self.mappings = mappings

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    results = [mapping(inputs) for mapping in self.mappings]
    return jax.tree.map(lambda *args: sum(args), *results)

  @property
  def output_shapes(self):
    return self.mappings[0].output_shapes


class Embedding(nnx.Module):
  """Generates floating-value embeddings from inputs using a pytree mapping."""

  def __init__(
      self,
      output_shapes: dict[str, typing.ShapeDtypeStruct],
      feature_module: nnx.Module,
      mapping_factory: nnx.Module,
      transform: pytree_transforms.Transform = pytree_transforms.Identity(),
      input_state_shapes: typing.Pytree | None = None,
      *,
      rngs: nnx.Rngs,
  ):
    if input_state_shapes is None:
      input_state_shapes = MINIMAL_STATE_STRUCT  # default.
    self.feature_module = feature_module
    self.mapping = mapping_factory(
        input_shapes=feature_module.output_shapes(input_state_shapes),
        output_shapes=output_shapes,
        rngs=rngs,
    )
    self.transform = transform

  def __call__(self, inputs: typing.Pytree):
    # TODO(dkochkov): Figure out if we need to support sharding here, it would
    # require passing coords object as they are holding sharding information.
    # inputs = self.coords.with_dycore_sharding(inputs)
    features = self.feature_module(inputs)
    # features = self.coords.dycore_to_physics_sharding(features)
    outputs = self.mapping(features)
    # outputs = self.coords.physics_to_dycore_sharding(outputs)
    outputs = self.transform(outputs)
    # outputs = self.coords.with_dycore_sharding(outputs)
    return outputs

  @property
  def output_shapes(self) -> typing.Pytree:
    return self.transform.output_shapes(self.mapping.output_shapes)


class LandSeaIceEmbedding(nnx.Module):
  """Embedding module that combines embeddings over land, sea and sea ice."""

  def __init__(
      self,
      output_shapes: dict[str, typing.ShapeDtypeStruct],
      sea_embedding_factory,
      land_embedding_factory,
      sea_ice_embedding_factory,
      land_sea_mask_features,
      sea_ice_features,
      transform: pytree_transforms.Transform = pytree_transforms.Identity(),
      *,
      rngs: nnx.Rngs,
  ):
    self.land_embedding = land_embedding_factory(
        output_shapes=output_shapes, rngs=rngs
    )
    self.sea_embedding = sea_embedding_factory(
        output_shapes=output_shapes, rngs=rngs
    )
    self.sea_ice_embedding = sea_ice_embedding_factory(
        output_shapes=output_shapes, rngs=rngs
    )
    self.land_sea_mask_features = land_sea_mask_features
    self.sea_ice_features = sea_ice_features
    self.transform = transform
    self._output_shapes = output_shapes

  def __call__(
      self,
      inputs: typing.Pytree,
  ) -> typing.Pytree:
    """Returns the embedding output on nodal locations."""
    # get outputs from each model
    land_outputs = self.land_embedding(inputs)
    sea_outputs = self.sea_embedding(inputs)
    sea_ice_outputs = self.sea_ice_embedding(inputs)

    # prepare masks with fractional values in [0, 1]
    land_fraction = self.land_sea_mask_features(inputs)['land_sea_mask']
    sea_fraction = 1 - land_fraction
    sea_ice_fraction = self.sea_ice_features(inputs)['sea_ice_cover']
    # weight and combine outputs
    land_weight = land_fraction
    sea_ice_weight = sea_ice_fraction * sea_fraction  # ice covered sea
    sea_weight = (1 - sea_ice_fraction) * sea_fraction  # sea without ice

    def tree_scale(a, x):
      # Multiply leaves of `x` by `a`.
      return jax.tree.map(lambda y: a * y, x)

    outputs = jax.tree.map(
        lambda a, b, c: a + b + c,
        tree_scale(land_weight, land_outputs),
        tree_scale(sea_weight, sea_outputs),
        tree_scale(sea_ice_weight, sea_ice_outputs),
    )
    return self.transform(outputs)

  @property
  def output_shapes(self) -> typing.Pytree:
    parts = [self.land_embedding, self.sea_embedding, self.sea_ice_embedding]
    shapes = [x.output_shapes for x in parts]
    sample_shape = shapes[0]
    for shape in shapes:
      if shape != sample_shape:
        raise ValueError(f'Inconsistent embedding output shapes: {shapes=}')
    return self.transform.output_shapes(sample_shape)


class CoordsStateMapping(nnx.Module):
  """Predicts a pytree of a state specified by coordinates.."""

  def __init__(
      self,
      *,
      coords: coordinates.DinosaurCoordinates,
      surface_field_names: tuple[str, ...],
      volume_field_names: tuple[str, ...],
      embedding_factory: PytreeMappingFactory,
      transform: pytree_transforms.Transform = pytree_transforms.Identity(),
      rngs: nnx.Rngs,
  ):
    output_shapes = {}
    for name in volume_field_names:
      output_shapes[name] = typing.ShapeFloatStruct(coords.shape)
    for name in surface_field_names:
      output_shapes[name] = typing.ShapeFloatStruct(
          (1,) + coords.horizontal.shape
      )
    self.coords = coords
    self.embedding = embedding_factory(output_shapes, rngs=rngs)
    self.transform = transform

  def __call__(self, inputs: typing.PyTreeState) -> typing.PyTreeState:
    output = self.embedding(inputs)
    output = self.transform(output)
    # outputs = self.coords.with_dycore_sharding(outputs)
    return output

  @property
  def output_shapes(self) -> typing.Pytree:
    return self.transform.output_shapes(self.embedding.output_shapes)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for encoders.py."""

import functools
from absl.testing import absltest
from absl.testing import parameterized
from dinosaur import coordinate_systems
from dinosaur import primitive_equations
from dinosaur import scales
from dinosaur import sigma_coordinates
from dinosaur import spherical_harmonic
from dinosaur import typing
from dinosaur import vertical_interpolation
from dinosaur import xarray_utils
import haiku as hk
import jax
from neuralgcm import encoders
from neuralgcm import mappings
from neuralgcm import towers
from neuralgcm import transforms
import numpy as np


ModelState = typing.ModelState
units = scales.units


class TowerMock(hk.Module):
  """Vertical tower mock that that returns random outputs of expected shape."""

  def __init__(self, output_size, *args):
    super().__init__(name=None)
    self.output_size = output_size

  def __call__(self, inputs):
    del inputs  # unused.
    return jax.random.uniform(jax.random.PRNGKey(42), shape=[self.output_size])


class FeaturesMock(hk.Module):
  """Mock for features module that returns random outputs of expected shape."""

  def __init__(self, coords, *args):
    super().__init__(name=None)
    self.feature_shape = coords.nodal_shape
    self.key = jax.random.PRNGKey(42)

  def __call__(self, inputs, next_inputs=None, forcing=None, randomness=None):
    del inputs, next_inputs, forcing, randomness  # unused.
    return {'rng_field': jax.random.uniform(self.key, shape=self.feature_shape)}


class EncodersTest(parameterized.TestCase):
  """Tests encoder modules."""

  @parameterized.named_parameters(
      dict(
          testcase_name='WeatherbenchToPrimitiveEncoder',
          encoder_module=encoders.WeatherbenchToPrimitiveEncoder,
          encoder_kwargs={}),
      dict(
          testcase_name='LearnedWeatherbenchToPrimitiveEncoder',
          encoder_module=encoders.LearnedWeatherbenchToPrimitiveEncoder,
          encoder_kwargs={
              'modal_to_nodal_model_features_module': FeaturesMock,
              'modal_to_nodal_data_features_module': FeaturesMock,
              'nodal_mapping_module': functools.partial(
                  mappings.NodalMapping, tower_factory=functools.partial(
                      towers.ColumnTower, column_net_factory=TowerMock)),
              'correction_transform_module': transforms.IdentityTransform,
              'prediction_mask':
                  {'divergence': True, 'vorticity': True,
                   'log_surface_pressure': True,
                   'temperature_variation': True, 'sim_time': False,
                   'tracers': {}},
          },
      ),
  )
  def test_weatherbench_to_primitive_encoder(
      self,
      encoder_module,
      encoder_kwargs,
  ):
    """Tests that encoder produces expected output structures."""
    n_sigma_layers = 18
    pressure_levels = [50, 100, 500, 700, 1000]
    input_coords = coordinate_systems.CoordinateSystem(
        spherical_harmonic.Grid.T21(),
        vertical_interpolation.PressureCoordinates(pressure_levels),)
    coords = coordinate_systems.CoordinateSystem(
        spherical_harmonic.Grid.T21(),
        sigma_coordinates.SigmaCoordinates.equidistant(n_sigma_layers)
    )
    physics_specs = primitive_equations.PrimitiveEquationsSpecs.from_si()
    dt = physics_specs.nondimensionalize(10 * units.minute)
    aux_features = {
        xarray_utils.OROGRAPHY: np.ones(input_coords.horizontal.nodal_shape),
        xarray_utils.REF_TEMP_KEY: np.ones(coords.vertical.layers) * 288,
    }

    def encode_fwd(inputs):
      encoder = encoder_module(
          coords, dt, physics_specs, aux_features, input_coords,
          **encoder_kwargs)
      return encoder(inputs, forcing={})
    encoder_model = hk.transform(encode_fwd)

    inputs = {
        'u': np.ones(input_coords.modal_shape),
        'v': np.ones(input_coords.modal_shape),
        'z': np.ones(input_coords.modal_shape),
        't': np.ones(input_coords.modal_shape),
        'sim_time': np.asarray(0.0),
        'tracers': {},
    }
    inputs = jax.tree.map(lambda x: x[np.newaxis, ...], inputs)
    rng = jax.random.PRNGKey(42)
    params = encoder_model.init(rng, inputs)
    encoded_state = (encoder_model.apply(params, rng, inputs).state).asdict()

    get_shape_fn = lambda tree: jax.tree.map(np.shape, tree)
    expected_shapes = primitive_equations.StateWithTime(
        divergence=coords.modal_shape,
        vorticity=coords.modal_shape,
        temperature_variation=coords.modal_shape,
        log_surface_pressure=coords.surface_modal_shape,
        tracers=jax.tree.map(
            lambda x: coords.modal_shape,
            inputs['tracers']
        ),
        sim_time=tuple(),
    ).asdict()
    self.assertSameStructure(get_shape_fn(encoded_state), expected_shapes)


if __name__ == '__main__':
  jax.config.parse_flags_with_absl()
  absltest.main()
